import importlib
from pathlib import Path
import yaml
from .functions import Function

registered_functions = {}

with importlib.resources.open_text('charlib.characterizer.functions', 'functions.yml') as file:
    functions = yaml.safe_load(file)
    for name in functions:
        registered_functions[name] = Function(**functions[name])



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse, os, yaml
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from liberty.parser import parse_liberty
from PySpice.Logging import Logging

from charlib.characterizer.Characterizer import Characterizer
from charlib.characterizer.functions.functions import generate_yml


def main():
    """Run CharLib"""
    # Set up charlib arguments
    parser = argparse.ArgumentParser(
            prog='charlib',
            description='Characterize combinational and sequential standard cells.')
    parser.add_argument('--debug', action='store_true',
            help='Display extra information useful for debugging')
            
    # Set up run, compare, and generate_functions subcommands
    subparser = parser.add_subparsers(title='subcomamands', required=True)
    parser_characterize = subparser.add_parser('run', help='Characterize a cell library')
    parser_compare = subparser.add_parser('compare', help='Compare two liberty files')
    parser_genfunctions = subparser.add_parser('generate_functions', help='Generate YAML maps for registered functions')

    # Set up charlib run arguments
    parser_characterize.add_argument('library', type=str,
            help='The directory containing the library characterization configuration file')
    parser_characterize.add_argument('--multithreaded', action='store_true',
            help='Enable multithreaded execution')
    parser_characterize.add_argument('--comparewith', type=str, default='',
            help='A liberty file to compare results with.')
    parser_characterize.set_defaults(func=run_charlib)

    # Set up charlib compare arguments
    def compare_helper(args):
        """Helper function for compare subcommand"""
        with open(Path(args.compared), 'r') as compared:
            compare(args.benchmark, compared.read())
    parser_compare.add_argument('benchmark',  type=str,
            help='A liberty file to use as a benchmark for comparison')
    parser_compare.add_argument('compared', type=str,
            help='A liberty file to compare against the benchmark')
    parser_compare.set_defaults(func=compare_helper)

    # Set up charlib generate_functions arguments
    def genfunctions_helper(args):
        """Helper function for generate_functions subcommand"""
        generate_yml()
    # TOOD: Add argument for expression map
    parser_genfunctions.set_defaults(func=genfunctions_helper)

    # Parse args and execute
    args = parser.parse_args()
    args.func(args)


def run_charlib(args):
    """Run characterization and return the library"""
    library_dir = args.library
    
    # Search for a YAML file with the required config information
    print(f'Searching for YAML files in {str(library_dir)}')
    config = None
    for file in Path(library_dir).rglob('*.yml'):
        try:
            with open(file, 'r') as f:
                config = yaml.safe_load(f)
                f.close()
        except yaml.YAMLError as e:
            print(e)
            print(f'Skipping "{str(file)}": file contains invalid YAML')
            continue
        if config.keys() >= {'settings', 'cells'}:
            break # We have found a YAML file with config information
    if not config:
        raise FileNotFoundError(f'Unable to locate a YAML file containing configuration settings in {library_dir} or its subdirectories.')
    print(f'Reading configuration found in "{str(file)}"')

    # Override settings with command line settings
    settings = config['settings']
    settings['debug'] = args.debug
    settings['multithreaded'] = args.multithreaded
    cells = config['cells']

    # Read in library settings
    characterizer = Characterizer(**settings)
    logger = Logging.setup_logging(logging_level='ERROR')

    # Read cells
    for name, properties in cells.items():
        # If properties is a (name, filepath) pair, fetch cell config from YAML at filepath
        if isinstance(properties, str):
            # Search within library_dir for the specified file
            for file in Path(library_dir).rglob(properties):
                with open(file, 'r') as f:
                    properties = yaml.safe_load(f)
                    f.close()
                    break # Quit searching after successfully reading a match

        # Merge settings.cell_defaults into properties, keeping entries from properties
        for key, value in characterizer.settings.cell_defaults.items():
            if not key in properties.keys():
                properties[key] = value

        # Read config data for this cell
        inputs = properties.pop('inputs')
        outputs = properties.pop('outputs')
        functions = properties.pop('functions')
        clock = properties.pop('clock', None)
        flops = properties.pop('flops', [])

        # Add cells
        if clock:
            characterizer.add_flop(name, inputs, outputs, clock, flops, functions, **properties)
        else:
            characterizer.add_cell(name, inputs, outputs, functions, **properties)

    # TODO: Add print statements to display which keys in YAML were not used

    # Characterize
    library = characterizer.characterize()

    # Export
    results_dir = characterizer.settings.results_dir
    results_dir.mkdir(parents=True, exist_ok=True)
    libfile_name = results_dir / f'{library.name}.lib'
    with open(libfile_name, 'w') as libfile:
        libfile.write(str(library))
        print(f'Results written to {str(libfile_name.resolve())}')

    # Run any post-characterization analysis
    if args.comparewith:
        compare(args.comparewith, library)


def compare(benchmark, characterized):
    """Compare prop delay and trans delay for each cell and make scatter plots"""
    charlib_rise_prop_data = []
    benchmark_rise_prop_data = []
    charlib_fall_prop_data = []
    benchmark_fall_prop_data = []
    charlib_rise_trans_data = []
    benchmark_rise_trans_data = []
    charlib_fall_trans_data = []
    benchmark_fall_trans_data = []
    
    benchmark_lib = parse_liberty(open(Path(benchmark), 'r').read())
    characterized_lib = parse_liberty(str(characterized))

    # Iterate over cells
    for charlib_cell in characterized_lib.get_groups('cell'):
        # Find this cell in the benchmark lib
        benchmark_cell = None
        for cell in benchmark_lib.get_groups('cell'):
            if cell.args[0] == charlib_cell.args[0]:
                benchmark_cell = cell
                break
        if benchmark_cell is None:
            print(f'Skipping cell "{charlib_cell.args[0]}": not found in {str(benchmark)}.')
            continue # Skip to next cell

        # Iterate over pins
        for charlib_pin in charlib_cell.get_groups('pin'):
            # Find the pin on the benchmark cell
            benchmark_pin = None
            for pin in benchmark_cell.get_groups('pin'):
                if pin.args[0] == charlib_pin.args[0]:
                    benchmark_pin = pin
                    break
            if benchmark_pin is None:
                print(f'Skipping pin "{charlib_cell.args[0]}.{charlib_pin.args[0]}": not found in {str(benchmark)}')
                continue # Skip to next pin

            # Iterate over timings
            for charlib_timing in charlib_pin.get_groups('timing'):
                # Find the timing on the benchmark pin
                benchmark_timing = None
                for timing in benchmark_pin.get_groups('timing'):
                    if timing['related_pin'] == charlib_timing['related_pin']:
                        benchmark_timing = timing
                        break
                if benchmark_timing is None:
                    print(f'Skipping timing for {charlib_cell.args[0]}.{charlib_pin.args[0]}.{charlib_timing["related_pin"]}')
                    continue

                # Average propagation delay
                charlib_rise_prop_data.append(np.mean(charlib_timing.get_groups('cell_rise')[0].get_array('values')))
                benchmark_rise_prop_data.append(np.mean(benchmark_timing.get_groups('cell_rise')[0].get_array('values')))
                charlib_fall_prop_data.append(np.mean(charlib_timing.get_groups('cell_fall')[0].get_array('values')))
                benchmark_fall_prop_data.append(np.mean(benchmark_timing.get_groups('cell_fall')[0].get_array('values')))

                # Average transient delay
                charlib_rise_trans_data.append(np.mean(charlib_timing.get_groups('rise_transition')[0].get_array('values')))
                benchmark_rise_trans_data.append(np.mean(benchmark_timing.get_groups('rise_transition')[0].get_array('values')))
                charlib_fall_trans_data.append(np.mean(charlib_timing.get_groups('fall_transition')[0].get_array('values')))
                benchmark_fall_trans_data.append(np.mean(benchmark_timing.get_groups('fall_transition')[0].get_array('values')))

    # Plot propagation delay data
    prop_figure, prop_ax = plt.subplots()
    prop_ax.grid()
    print(benchmark_rise_prop_data, charlib_rise_prop_data)
    prop_ax.scatter(benchmark_rise_prop_data, charlib_rise_prop_data, label='Rising')
    prop_ax.scatter(benchmark_fall_prop_data, charlib_fall_prop_data, label='Falling')
    prop_limits = [0, 1.1*max(max(benchmark_rise_prop_data), max(charlib_rise_prop_data))]
    prop_ax.plot(prop_limits, prop_limits, color='black', alpha=0.2, label='Ideal')
    prop_ax.set(
        xlim=prop_limits,
        ylim=prop_limits,
        xlabel='Benchmark Propagation Delay [ns]',
        ylabel='CharLib Propagation Delay [ns]',
        title='Propagation Delay Correlation'
    )
    prop_ax.legend()

    # Plot transient delay data
    tran_figure, tran_ax = plt.subplots()
    tran_ax.grid()
    tran_ax.scatter(benchmark_rise_trans_data, charlib_rise_trans_data, label='Rising')
    tran_ax.scatter(benchmark_fall_trans_data, charlib_fall_trans_data, label='Falling')
    tran_limits = [0, 1.1*max(max(benchmark_rise_trans_data), max(charlib_rise_trans_data))]
    tran_ax.plot(tran_limits, tran_limits, color='black', alpha=0.2, label='Ideal')
    tran_ax.set(
        xlim=tran_limits,
        ylim=tran_limits,
        xlabel='Benchmark Transient Delay [ns]',
        ylabel='CharLib Transient Delay [ns]',
        title='Transient Delay Correlation'
    )
    tran_ax.legend()

    plt.show()

    # Calculate absolute error
    rise_prop_ae = np.abs(np.asarray(charlib_rise_prop_data) - np.asarray(benchmark_rise_prop_data))
    fall_prop_ae = np.abs(np.asarray(charlib_fall_prop_data) - np.asarray(benchmark_fall_prop_data))
    rise_tran_ae = np.abs(np.asarray(charlib_rise_trans_data) - np.asarray(benchmark_rise_trans_data))
    fall_tran_ae = np.abs(np.asarray(charlib_fall_trans_data) - np.asarray(benchmark_fall_trans_data))
    # Calculate worst case absolute error
    rise_prop_max_ae = max(rise_prop_ae)
    fall_prop_max_ae = max(fall_prop_ae)
    rise_tran_max_ae = max(rise_tran_ae)
    fall_tran_max_ae = max(fall_tran_ae)
    # Calculate mean absolute error
    rise_prop_mae = np.mean(rise_prop_ae)
    fall_prop_mae = np.mean(fall_prop_ae)
    rise_tran_mae = np.mean(rise_tran_ae)
    fall_tran_mae = np.mean(fall_tran_ae)

    print('Rise prop', rise_prop_max_ae, rise_prop_mae)
    print('Fall prop', fall_prop_max_ae, fall_prop_mae)
    print('Rise tran', rise_tran_max_ae, rise_tran_mae)
    print('Fall tran', fall_tran_max_ae, fall_tran_mae)

if __name__ == '__main__':
    main()



# Token definitions 
T_NOT = 0
T_GROUP = 1
T_GROUP_END = 2
T_AND = 3
T_XOR = 4
T_OR = 5
T_OTHER = 6

class Token:
    def __init__(self, symbol: str) -> None:
        self.symbol = symbol if not symbol == '!' else '~'
        self.type = symbol

    def __str__(self) -> str:
        return self.symbol
    
    def __repr__(self) -> str:
        return f'Token("{self.symbol}")'
    
    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.symbol == other.symbol
    
    @property
    def type(self) -> int:
        return self._type

    @type.setter
    def type(self, value: str):
        if value == '~' or value == '!':
            self._type = T_NOT
        elif value == '(':
            self._type = T_GROUP
        elif value == ')':
            self._type = T_GROUP_END
        elif value == '&':
            self._type = T_AND
        elif value == '^':
            self._type = T_XOR
        elif value == '|':
            self._type = T_OR
        else:
            self._type = T_OTHER

# Rules
RULES = lambda S : [
    [Token('~'), 'E', 'O'],             # 0: E -> ~EO
    [Token('('), 'E', Token(')'), 'O'], # 1: E -> (E)O
    [S, 'O'],                           # 2: E -> [A-z,0-9,_]*O
    [Token('&'), 'E', 'O'],             # 3: O -> &EO
    [Token('^'), 'E', 'O'],             # 4: O -> ^EO
    [Token('|'), 'E', 'O'],             # 5: O -> |EO
    [],                                 # 6: O -> null
]

def _get_rule(stack_token, input_token) -> int:
    """Look up a rule by the current stack token and input token"""
    # The left column is the current stack token and the top row is the
    # current input token. The cell selected by these two tokens is the
    # rule we will use to resolve the stack token. Empty cells should
    # result in errors.
    # |     | '~' | '(' | '&' | '^' | '|' | anything else |
    # | --- | --- | --- | --- | --- | --- | ------------- |
    # | 'E' |  0  |  1  |     |     |     |       2       |
    # | 'O' |     |     |  3  |  4  |  5  |       6       |
    rule = 999 # Default to invalid rule
    if stack_token == 'E':
        if input_token.type == T_NOT:
            rule = 0
        elif input_token.type == T_GROUP:
            rule = 1
        elif input_token.type == T_OTHER:
            rule = 2
    elif stack_token == 'O':
        if input_token.type == T_AND:
            rule = 3
        elif input_token.type == T_XOR:
            rule = 4
        elif input_token.type == T_OR:
            rule = 5
        elif input_token.type == T_OTHER or input_token.type == T_GROUP_END:
            rule = 6
    return rule

def _parse(tokens: list) -> list:
    stack = ['O', 'E']
    position = 0
    rule_sequence = []
    while stack:
        stack_token = stack.pop()
        if isinstance(stack_token, Token):
            # Pop resolved terminals
            input_token = tokens[position]
            if stack_token == input_token:
                position += 1
            else:
                raise ValueError(f'Failed to parse logic string "{"".join([str(t) for t in tokens])}". Parsing failed at token "{input_token.symbol}" (position {position})')
        else:
            # Resolve rules using lookup table
            try:
                input_token = tokens[position]
            except IndexError:
                input_token = Token('') # Empty token of type T_OTHER
            rule = _get_rule(stack_token, input_token)
            try:
                stack.extend(reversed(RULES(input_token)[rule]))
            except IndexError: # Error due to nonexistent rule
                raise ValueError(f'Failed to parse logic string "{"".join([str(t) for t in tokens])}". Parsing failed at position {position}')
            rule_sequence.append(rule)
    
    # Now that we know the production rules, produce an AST in prefix format
    syntax_tree = []
    named_tokens = [n for n in tokens if n.type == T_OTHER]
    for rule in rule_sequence:
        if rule == 0:
            syntax_tree.extend(['O', '~'])
        elif rule == 1:
            syntax_tree.append('O')
        elif rule == 2:
            syntax_tree.extend(['O', named_tokens.pop(0).symbol])
        elif rule == 3:
            last_O_index = -1 - [t for t in reversed(syntax_tree)].index('O')
            syntax_tree.insert(last_O_index+1, '&')
        elif rule == 4:
            last_O_index = -1 - [t for t in reversed(syntax_tree)].index('O')
            syntax_tree.insert(last_O_index+1, '^')
        elif rule == 5:
            last_O_index = -1 - [t for t in reversed(syntax_tree)].index('O')
            syntax_tree.insert(last_O_index+1, '|')
        elif rule == 6:
            try:
                last_O_index = -1 - [t for t in reversed(syntax_tree)].index('O')
                syntax_tree.pop(last_O_index)
            except ValueError:
                pass # We don't add O as aggressively in tree generation, so it's ok if it isn't present
    return syntax_tree

def _lex(expression: str) -> list:
    """Convert a simple boolean logic expression into tokens"""
    tokens = []
    temp = ''
    for c in ''.join(expression.split()):
        if c in ['~', '!', '(', ')', '|', '&', '^']:
            if temp:
                tokens.append(Token(temp))
                temp = ''
            tokens.append(Token(c))
        elif c.isalnum() or c == '_':
            temp += c
        else:
            raise ValueError(f'Invalid character in boolean expression: {c}')
    if temp:
        tokens.append(Token(temp))
    return tokens

def parse_logic(expression: str) -> list:
    """Parse a logic string and return the result in Polish prefix notation.
    
    This parser supports parentheses, NOT (~), AND (&), OR (|), and XOR (^) operations.
    The return value is a list of tokens in Polish prefix notation.

    Note that this parser does not currently support operator precedence. If you want
    operations to happen in a particular order, make sure to use parentheses. 
    """
    return _parse(_lex(expression))

def _resolve_unates(syntax_tree: list, target: str):
    op = syntax_tree.pop(0)
    if op == '~':
        unate_l = -1
        unate_r = None
    elif op == '&':
        unate_l = 1
        unate_r = 1
    elif op == '|' or op == '^':
        # can't determine these yet - have to check which side contains the target
        unate_l = 0
        unate_r = 0
    else:
        return syntax_tree, {op: 1} # Return symbol
    # Resolve left side
    syntax_tree, unates = _resolve_unates(syntax_tree, target)
    # Check for target in left side
    if unate_l == 0 and target in unates:
        unate_l = 1
    else:
        unate_l = -1
    for k,u in unates.items():
        unates[k] = unate_l * u
    # Resolve right side
    if unate_r is not None:
        syntax_tree, unates_r = _resolve_unates(syntax_tree, target)
        # Check for target in right side
        if unate_r == 0 and target in unates_r:
            unate_r = 1
        else:
            unate_r = -1
        for k,u in unates_r.items():
            unates[k] = unate_r * u
    return syntax_tree, unates

def generate_test_vectors(expression: str, inputs: list) -> list:
    # This method doesn't work correctly for some expressions, and will be revisited and fixed later.
    test_vectors = []
    for target_port in inputs:
        # Generate two test vectors for each input: one rising, one falling
        for state in ['01', '10']:
            input_vector = []
            for port in inputs:
                # Get unates for each variable
                _, unates = _resolve_unates(parse_logic(expression), target_port)
                # Make sure we don't have any unexpected ports
                for k in unates.keys():
                    if k not in inputs:
                        raise ValueError(f'Unexpected port name {k} encountered during test vector generation')
                # Use unates to determine port states
                if port is target_port:
                    input_vector.append(state)
                    output_state = state if unates[port] > 0 else state[::-1]
                else:
                    input_vector.append(str(int(unates[port] > 0)))
            test_vectors.append([output_state, input_vector])
    return test_vectors


if __name__ == '__main__':
    # If run as main, test parser
    assert parse_logic('~(A^B&C)') == ['~', '^', 'A', '&', 'B', 'C']
    assert parse_logic('_^B | potato') == ['^', '_', '|', 'B', 'potato']
    assert parse_logic('~~~~A') == ['~', '~', '~', '~', 'A']
    assert parse_logic('(~(A&~C)) ^ B') == ['^', '~', '&', 'A', '~', 'C', 'B']
    assert parse_logic('A&B&C&D&E&F&G&H&I&J&K') == ['&', 'A', '&', 'B', '&', 'C', '&', 'D', '&', 'E', '&', 'F', '&', 'G', '&', 'H', '&', 'I', '&', 'J', 'K']
    tokens = _lex('~&^|')
    assert tokens == [Token('~'), Token('&'), Token('^'), Token('|')]
    try:
        _parse(tokens) # We expect this to fail
        raise AssertionError # This line should never be executed
    except ValueError:
        pass # Pass if we catch a ValueError
    assert parse_logic('b&a&a') == ['&', 'b', '&', 'a', 'a']
    assert parse_logic('(C&(A^B))|(A&B)') == ['|','&','C','^','A','B','&','A','B']
    print(generate_test_vectors('~A', ['A']))


"""This module contains data structures used to read and write liberty files"""

from dataclasses import dataclass

import numpy as np
import matplotlib.pyplot as plt

class Cell:
    """A single standard cell"""
    def __init__(self, name: str, area: int=0, **attrs) -> None:
        """Create a new cell

        :param name: cell name
        :param area: cell area
        :param **attrs: additional liberty attributes, specified as key-value pairs"""

        self._name = name.upper()
        self.area = area
        self._attrs = attrs
        self.flops = []
        self.pins = {}

    @classmethod
    def from_str(value: str):
        """Parse a cell from a liberty string"""
        pass # TODO

    @property
    def name(self) -> str:
        """Return cell name"""
        return self._name

    def __getitem__(self, key: str):
        """Return self[key]. Searches pins by name."""
        return self.pins[key.upper()]

    def add_ff(self, in_node: str, out_node: str, next_state: str, clock: str, **attrs):
        """Add a new flop to this cell's list of flops"""
        ff = Flop(in_node, out_node, next_state, clock)
        for attr, value in attrs:
            setattr(ff, attr, value)
        self.flops.append()

    def add_pin(self, name: str, direction=None, role='io'):
        """Add a new pin to this cell's list of pins"""
        name = name.upper()
        self.pins[name] = Pin(name, direction, role)

    @property
    def attributes(self) -> dict:
        """Return extra liberty attributes for this cell"""
        return self._attrs

    def add_attribute(self, key: str, value: str):
        """Add a new liberty attribute to this cell"""
        self._attrs[key] = str(value)

    def is_pad_cell(self):
        """Return `True` if this cell contains any pad pins."""
        return 'pad' in [pin.role for pin in self.pins.values()]

    def templates(self) -> list:
        """Generate templates for all lookup tables in this cell's pins"""
        templates = []
        for pin in self.pins.values():
            templates.extend(pin.templates())
        return list(set(templates))

    def __str__(self) -> str:
        """Return str(self)"""
        lib_str = [f'cell ({self.name}) {{']
        if 'BUF' in self.name.upper():
            lib_str.append('  cell_footprint : buf;')
        elif 'INV' in self.name.upper():
            lib_str.append('  cell_footprint : inv;')
        lib_str.append(f'  area : {self.area};')
        if self.is_pad_cell():
            lib_str.append('  pad_cell : true;')
        for key, value in self.attributes:
            lib_str.append(f'  {key} : {value};')
        for flop in self.flops:
            for line in str(flop).split('\n'):
                lib_str.append(f'  {line}')
        for pin in dict(sorted(self.pins.items())).values():
            for line in str(pin).split('\n'):
                lib_str.append(f'  {line}')
        lib_str.append('}')
        return '\n'.join(lib_str)


class Flop:
    """A single flip-flop internal to a standard cell"""
    def __init__(self, in_name, out_name, next_state, clock):
        self.in_name = in_name
        self.out_name = out_name
        self.next_state = next_state
        self.clocked_on = clock
        self.clear = ''
        self.preset = ''
        self.clear_preset_var1 = 'L'

    def __str__(self) -> str:
        """Return str(self)"""
        ff_str = [
            f'ff ({self.in_name}, {self.out_name}) {{',
            f'  next_state : "{self.next_state}";',
            f'  clocked_on : "{self.clocked_on}";',
        ]
        if self.clear:
            ff_str.append(f'  clear : "{self.clear}";')
        if self.preset:
            ff_str.append(f'  preset : "{self.preset}";')
        if self.clear_preset_var1:
            ff_str.append(f'  clear_preset_var1 : {self.clear_preset_var1};')
        ff_str.append('}')
        return '\n'.join(ff_str)


class Pin:
    """A single pin from a standard cell"""
    def __init__(self, name: str, direction: str|None=None, role: str='io') -> None:
        """Create a new Pin.

        :param name: pin name
        :param direction: (optional) pin direction
        :param role: (optional) pin role
        """
        self.capacitance = 0
        self.drive_current = 0
        self.rise_capacitance = 0
        self.fall_capacitance = 0
        self.max_capacitance = 0
        self.function = ''
        self.three_state = ''
        self.internal_power = []
        self.min_pulse_width_high = 0
        self.min_pulse_width_low = 0
        self.timings = []

        self._name = name.upper()
        if direction in ['input', 'output', None]:
            self._direction = direction
        else:
            raise ValueError('Pin direction must be one of ["input", "output", None]')
        if role in ['set', 'reset', 'clock', 'io', 'pad']:
            self._role = role
        else:
            raise ValueError('Pin role must be one of ["set", "reset", "clock", "io", "pad"]')

    @property
    def name(self) -> str:
        """Return pin name."""
        return self._name

    @property
    def direction(self) -> str:
        """Return pin direction."""
        return self._direction

    @property
    def role(self) -> str:
        """Return pin role. May be 'set', 'reset', 'clock', or 'io'."""
        return self._role

    def is_clk(self) -> bool:
        """Return `True` if the pin's role is `'clock'`"""
        return self.role == 'clock'

    def is_set(self) -> bool:
        """Return `True` if the pin's role is `'set'`"""
        return self.role == 'set'

    def is_reset(self) -> bool:
        """Return `True` if the pin's role is `'reset'`"""
        return self.role == 'reset'

    def is_io(self) -> bool:
        """Return `True` if the pin's role is `'io'`"""
        return self.role == 'io'

    def is_pad(self) -> bool:
        """Return `True` if the pin's role is `'pad'`"""
        return self.role == 'pad'

    def __eq__(self, other) -> bool:
        """Return `True` if name, role, and direction match.
        
        :param other: A pin to compare against self"""
        return self.name == other.name \
            and self.role == other.role \
                and self.direction == other.direction

    def __str__(self) -> str:
        """Return str(self)"""
        # Pin properties
        lib_str = [f'pin ({self.name}) {{']
        if self.is_pad():
            lib_str.append('  is_pad : true;')
        lib_str.append(f'  direction : {self.direction};')
        if self.is_pad() and self.drive_current:
            lib_str.append(f'  drive_current : {self.drive_current:7f};')
        lib_str.extend([
            f'  capacitance : {self.capacitance:7f};',
            f'  rise_capacitance : {self.rise_capacitance:7f};',
            f'  fall_capacitance : {self.fall_capacitance:7f};',
        ])
        if self.is_clk():
            lib_str.append('  clock : true;')
        if self.max_capacitance:
            lib_str.append(f'  max_capacitance : {self.max_capacitance:7f};')
        if self.function:
            lib_str.append(f'  function : "{self.function}";')
        if self.three_state:
            lib_str.append(f'  three_state : "{self.three_state}";')
        # Internal power
        for data in self.internal_power:
            for line in str(data).split('\n'):
                lib_str.append(f'  {line}')
        # min_pulse_width
        if self.role in ['clock', 'set', 'reset']:
            if self.min_pulse_width_high:
                lib_str.append(f'  min_pulse_width_high : {self.min_pulse_width_high};')
            if self.min_pulse_width_low:
                lib_str.append(f'  min_pulse_width_low : {self.min_pulse_width_low};')
        # Timing
        for data in self.timings:
            for line in str(data).split('\n'):
                lib_str.append(f'  {line}')
        lib_str.append('}')
        return '\n'.join(lib_str)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f'Pin(name="{self.name}", direction="{self.direction}", role="{self.role}")'

    def plot_delay(self, settings, cell_name):
        """Plot propagation and transient delays for rise and fall cases

        Generate two plots: one for rise timing and one for fall timing. Both plots
        should display propagation and transient delay as a function of slew rate and
        capacitive load."""
        # Input pins may not have delay data to plot. Prevent plotting in these cases
        if self.direction == 'output':
            figs = []
            # Output pins have cell_{rise,fall} and {rise,fall}_transition
            for timing_data in self.timings:
                for direction in ['rise', 'fall']:
                    fig = plt.figure(
                        label=f'{cell_name} | {self.name} ({direction}) Timing | Related Pin: {timing_data.related_pin}'
                    )

                    prop_table = timing_data[f'cell_{direction}']
                    tran_table = timing_data[f'{direction}_transition']

                    ax, indices = prop_table.generate_axes(fig)
                    prop_surface = ax.plot_surface(*indices, np.asarray(prop_table.data()), edgecolor='red', cmap='inferno', alpha=0.3, label='Propagation Delay')
                    prop_surface._edgecolors2d = prop_surface._edgecolor3d # Workaround for legend. See https://stackoverflow.com/questions/54994600/pyplot-legend-poly3dcollection-object-has-no-attribute-edgecolors2d
                    prop_surface._facecolors2d = prop_surface._facecolor3d # Workaround for legend
                    tran_surface = ax.plot_surface(*indices, np.asarray(tran_table.data()), edgecolor='blue', cmap='viridis', alpha=0.3, label='Transient Delay')
                    tran_surface._edgecolors2d = tran_surface._edgecolor3d # Workaround for legend.
                    tran_surface._facecolors2d = tran_surface._facecolor3d # Workaround for legend
                    time_unit = str(settings.units.time.prefixed_unit)
                    cap_unit = str(settings.units.capacitance.prefixed_unit)
                    ax.set(
                        xlabel=f'Slew Rate [{time_unit}]',
                        ylabel=f'Fanout [{cap_unit}]',
                        zlabel=f'Delay [{time_unit}]',
                        title='Transient and Propagation Delays'
                    )
                    ax.legend()

                    figs.append(fig)
            return figs

    def plot_energy(self, settings):
        """Plot energy for rise and fall cases"""
        pass # TODO

    def templates(self) -> list:
        """Generate all lookup table templates required for this pin"""
        templates = []
        [templates.extend(timing.templates()) for timing in self.timings]
        [templates.extend(power.templates()) for power in self.internal_power]
        return list(set(templates))


class InternalPowerData:
    """Container for pin power tables"""
    def __init__(self, related_pin: str='') -> None:
        """Create a new InternalPowerData.

        Initialize a new blank InternalPowerData. fall_power and rise_power must be
        initialized using their corresponding set functions, set_fall_power_table and
        set_rise_power_table.

        :param related_pin: (optional) the related pin for the power data (if relevant)"""
        self._related_pin = related_pin
        self._rise_power = None
        self._fall_power = None

    @property
    def related_pin(self) -> str:
        """Return related pin name."""
        return self._related_pin

    @property
    def rise_power(self):
        """Return rise_power table."""
        return self._rise_power

    def set_rise_power_table(self, template: str, values: list, index_1: list, index_2: list):
        """Add rise_power table data"""
        self._rise_power = Table('rise_power', template, values, index_1, index_2)

    @property
    def fall_power(self):
        """Return fall_power table."""
        return self._fall_power

    def set_fall_power_table(self, template: str, values: list, index_1: list, index_2: list):
        """Set fall_power table data"""
        self._fall_power = Table('fall_power', template, values, index_1, index_2)

    def templates(self) -> list:
        return [table.template_str() for table in [self._rise_power, self._fall_power]]

    def __str__(self) -> str:
        """Return str(self)"""
        internal_power_str = ['internal_power () {']
        if self.related_pin:
            internal_power_str.append(f'  related_pin : "{self.related_pin}";')
        for line in str(self.rise_power).split('\n'):
            internal_power_str.append(f'  {line}')
        for line in str(self.fall_power).split('\n'):
            internal_power_str.append(f'  {line}')
        internal_power_str.append('}')
        return '\n'.join(internal_power_str)


class TimingData:
    """Container for pin timing data and tables"""
    def __init__(self, related_pin, timing_type='', **attrs) -> None:
        """Create a new TimingData"""
        self._related_pin = related_pin
        self._timing_type = timing_type
        self._attrs = attrs
        self._tables = {}

    @property
    def related_pin(self) -> str:
        """Return related pin name."""
        return self._related_pin

    @property
    def timing_type(self) -> str:
        """Return timing type"""
        return self._timing_type

    def add_table(self, name, template, values, index_1, index_2=None) -> None:
        """Add a new table to the timing data

        :param name: the table's name as it will appear in liberty syntax
        :param template: the table template to use
        :param values: a list of table values
        :param index_1: a list of table indexes for dimension 1
        :param index_2: (optional) a list of table indexes for dimension 2"""
        self._tables[name] = Table(name, template, values, index_1, index_2)

    @property
    def attributes(self) -> dict:
        """Get attributes on this timing data group"""
        return self._attrs

    def add_attribute(self, key, value):
        """Add a new attribute to this TimingData"""
        self._attrs[key] = value

    def templates(self) -> list:
        """Get table templates for this TimingData"""
        return [table.template_str() for table in self._tables.values()]

    def __getitem__(self, key):
        """Return self[key]. Searches tables by name."""
        return self._tables[key]

    def __str__(self) -> str:
        """Return str(self)."""
        timing_str = [
            'timing () {',
            f'  related_pin : "{self.related_pin}";',
        ]
        if self.timing_type:
            timing_str.append(f'  timing_type : {self.timing_type};')
        for key, value in self.attributes:
            timing_str.append(f'  {key} : {value};')
        for table in self._tables.values():
            for line in str(table).split('\n'):
                timing_str.append(f'  {line}')
        timing_str.append('}')
        return '\n'.join(timing_str)


@dataclass
class TableTemplate:
    """Template data for a Table"""
    name = 'delay_template'
    variables = ['input_net_transition']

    def __str__(self) -> str:
        """Return str(self)"""
        return self.name


class Table:
    """A Table contains tabular data as would be displayed in a liberty file."""
    def __init__(self, name: str, template: TableTemplate, values: list, index_1: list, index_2: list|None=None) -> None:
        """Create a new Table

        :param name: table name
        :param template: table template
        :param values: a list of values in the table
        :param index_1: indices used to lookup table values in dimension 1
        :param index_2: indices used to lookup table values in dimension 2 (if present)
        """
        self._name = name
        self._template = template
        self._index_1 = index_1
        self._index_2 = index_2

        if not len(values) == (len(index_1) if not self.is_2d() else len(index_1) * len(index_2)):
            raise ValueError('Incorrect number of values for supplied index_1/index_2')
        self._values = values

    @property
    def name(self) -> str:
        """Return table name"""
        return self._name

    @property
    def template(self) -> str:
        """Return table template data"""
        return self._template

    @property
    def values(self) -> list:
        """Return a list of values in the table"""
        return self._values

    @property
    def index_1(self) -> list:
        """Return indices used to lookup values in dimension 1"""
        return self._index_1

    @property
    def index_2(self) -> list:
        """Return indices used to lookup values in dimension 2"""
        return self._index_2

    def is_2d(self) -> bool:
        """Return `True` if the table has more than 1 row"""
        return bool(self.index_2)

    @property
    def shape(self):
        """Return the table shape (not including indices)."""
        return len(self.index_1) if not self.is_2d() else (len(self.index_1), len(self.index_2))

    def __str__(self) -> str:
        """Return str(self)"""
        table_str = [
            f'{self.name} ({self.template}) {{',
            f'  index_1 ("{", ".join(self.index_1)}");'
        ]
        if self.index_2:
            table_str.append(f'  index_2 ("{", ".join(self.index_2)}");')
        values = self.data().tolist()
        if self.is_2d():
            table_str.append('  values ( \\')
            rows = ', \\\n'.join([f'"{", ".join(group)}"' for group in values]).split('\n')
            for row in rows:
                table_str.append(f'    {row}')
            table_str[-1] += ');'
        else:
            table_str.append(f'  values ("{", ".join(values)}");')
        table_str.append('}')
        return '\n'.join(table_str)

    def data(self):
        """Return the table values reshaped according to the index lengths"""
        return np.reshape(self.values, self.shape)

    def generate_axes(self, figure):
        """Generate a set of axes appropriate for displaying this table's data"""
        to_float = lambda idx: [float(i) for i in idx]
        if self.is_2d():
            x_index = np.repeat(np.expand_dims(to_float(self.index_1), 1), len(self.index_2), 1)
            y_index = np.swapaxes(np.repeat(np.expand_dims(to_float(self.index_2), 1), len(self.index_1), 1), 0, 1)
            indices = [x_index, y_index]
            ax = figure.add_subplot(projection='3d')
            ax.set_proj_type('ortho')
        else:
            indices = to_float(self.index_1)
            ax = figure.add_subplot()

        return ax, indices

    def template_str(self):
        """Generate the template for this table"""
        template_str = [
            f'lu_table_template ({self.template.name}) {{',
            f'  variable_1 : {self.template.variables[0]};',
        ]
        if self.index_2:
            template_str.append(f'  variable_2 : {self.template.variables[1]};')
        template_str.append(f'  index_1 ("{", ".join(self.index_1)}");')
        if self.index_2:
            template_str.append(f'  index_2 ("{", ".join(self.index_2)}");')
        template_str.append('}')
        return '\n'.join(template_str)
        



"""Maps logic functions to truth tables and test vectors."""

import re

registered_expressions = {
    'BUF': 'a',
    'INV': '!a',
    'AND2': 'a&b',
    'AND3': 'a&b&c',
    'AND4': 'a&b&c&d',
    'OR2': 'a|b',
    'OR3': 'a|b|c',
    'OR4': 'a|b|c|d',
    'XOR2': 'a^b',
    'XOR3': 'a^b^c',
    'XOR4': 'a^b^c^d',
    'NAND2': '!(a&b)',
    'NAND3': '!(a&b&c)',
    'NAND4': '!(a&b&c&d)',
    'NOR2': '!(a|b)',
    'NOR3': '!(a|b|c)',
    'NOR4': '!(a|b|c|d)',
    'XNOR2': '!(a^b)',
    'XNOR3': '!(a^b^c)',
    'XNOR4': '!(a^b^c^d)',
    'AOI21': '!((a&b)|c)',
    'AOI22': '!((a&b)|(c&d))',
    'OAI21': '!((a|b)&c)',
    'OAI22': '!((a|b)&(c|d))',
    'SEL2': '(a&(!s))|(b&s)',
}


class Function:
    """Provides function evaluation and mapping faculties"""
    def __init__(self, expression: str, test_vectors: list=[]) -> None:
        """Initialize a new Function"""
        self.expression = expression.replace('!','~')
        self.stored_test_vectors = test_vectors

    @property
    def operands(self) -> list:
        """Return a list of operand names"""
        return list(set(re.findall(r'(\w+)', self.expression)))

    def eval(self, **inputs) -> bool:
        """Evaluate this function for the given inputs"""
        operands = self.operands
        if not len(inputs) == len(operands):
            raise ValueError(f'Expected {len(operands)} inputs for function {self.expression}, got {len(inputs)}')
        f = eval(f'lambda {",".join(operands)}: int({self.expression.replace("~", " not ")})')
        return f(**inputs)

    def truth_table(self) -> list:
        """Return a truth table for this function"""
        table = []
        length = len(self.operands)
        for n in range(2**length):
            input_vector = [int(c) for c in f'{n:0{length}b}']
            result = self.eval(**dict(zip(self.operands, input_vector)))
            table.append([input_vector, result])
        return table

    def __eq__(self, other) -> bool:
        """Compare two functions by checking that their truth tables are the same."""
        result = False
        try:
            result = self.truth_table() == other.truth_table()
        except AttributeError:
            # other is probably not a Function object; assume it's a str expression instead
            result = self == Function(other) # Recurse
        return result

    def __str__(self) -> str:
        """Return str(self)"""
        return self.expression.replace('~', '!')

    def __repr__(self) -> str:
        """Return repr(self)"""
        return f'Function("{self.expression}", {self.test_vectors})'

    def to_yaml(self, name) -> str:
        """Convert to a YAML-formatted string"""
        func_str = [
            f'{name}:',
            f'    expression: {self.expression}',
            '    test_vectors:'
        ]
        for tv in self.test_vectors:
            func_str.append(f'        - {tv}')
        return '\n'.join(func_str)

    @property
    def test_vectors(self) -> list:
        """Generate test vectors, or return stored configuration."""
        if self.stored_test_vectors: # If assigned and not empty, use stored
            return self.stored_test_vectors
        # The basic idea here is to use the truth tables we already have to generate test vectors.
        # Look at each row one at a time, and compare it to all lower rows in the truth table which
        # have a different output.
        # If a row differs only by a single input and the output, then we can use that delta for a
        # test vector.
        # TODO: Make this more efficient
        test_vectors = []
        table = self.truth_table()
        while table:
            row = table.pop(0) # Get top row of table
            # Compare to each later row with a differing output
            for compared_row in [r for r in table if not r[1] == row[1]]:
                # Check if input differs by only one pin
                delta_row = []
                delta_count = 0
                for i, j in zip(row[0], compared_row[0]):
                    if i == j:
                        delta_row.append(str(i))
                    else:
                        delta_row.append(f'{i}{j}')
                        delta_count += 1
                if delta_count == 1:
                    # Add the output to the new test vector
                    delta_row.append(f'{row[1]}{compared_row[1]}')
                    test_vectors.append(delta_row)
                    # Append a second copy with each element reversed
                    test_vectors.append([s[::-1] for s in delta_row])
        return test_vectors


def generate_yml():
    """Generates a YAML map of the registered expressions"""
    document = ""
    for name, expr in registered_expressions.items():
        func = Function(expr)
        document += func.to_yaml(name) + "\n"
    print(document)



"""Dispatches characterization jobs and manages cell data"""

from multiprocessing import Pool, cpu_count
from pathlib import Path

from charlib.liberty.UnitsSettings import UnitsSettings
from charlib.liberty.library import Library
from charlib.characterizer.TestManager import TestManager, CombinationalTestManager, SequentialTestManager

class Characterizer:
    """Main object of Charlib. Keeps track of settings and cells."""

    def __init__(self, **kwargs) -> None:
        self.settings = CharacterizationSettings(**kwargs)
        self.library = Library(kwargs.get('lib_name'), **kwargs)
        self.tests = []

    def add_cell(self, name, in_ports, out_ports, functions, **kwargs):
        """Create a new logic cell test"""
        self.tests.append(
            CombinationalTestManager(name, in_ports, out_ports, functions, **kwargs)
        )

    def add_flop(self, name, in_ports, out_ports, clock, flops, functions, **kwargs):
        """Create a new sequential cell test"""
        self.tests.append(
            SequentialTestManager(name, in_ports, out_ports, clock, flops, functions, **kwargs)
        )

    def characterize(self):
        """Characterize all cells"""

        # If no target cells were given, characterize all cells
        if self.settings.use_multithreaded:
            num_workers = max(len(self.tests), cpu_count())
            with Pool(num_workers) as pool:
                cells = pool.map(self.characterize_cell, [*self.tests])
        else:
            cells = [self.characterize_cell(cell) for cell in self.tests]

        # Add cells to the library
        [self.library.add_cell(cell) for cell in cells]

        return self.library

    def characterize_cell(self, cell):
        """Characterize a single cell. Helper for multiprocessing"""
        return cell.characterize(self.settings)


class CharacterizationSettings:
    """Container for characterization settings"""
    def __init__(self, **kwargs):
        """Create a new CharacterizationSettings instance"""
        # Behavioral settings
        self.simulator = kwargs.pop('simulator', 'ngspice-shared')
        self.use_multithreaded = kwargs.pop('multithreaded', True)
        self.results_dir = Path(kwargs.pop('results_dir', 'results'))
        self.debug = kwargs.pop('debug', False)
        self.debug_dir = Path(kwargs.pop('debug_dir', 'debug'))
        self.cell_defaults = kwargs.get('cell_defaults', {})

        # Units and important voltages
        self.units = UnitsSettings(**kwargs.get('units', {}))
        nodes = kwargs.pop('named_nodes', {})
        self.vdd = NamedNode(**nodes.get('vdd', {'name':'VDD'}))
        self.vss = NamedNode(**nodes.get('vss', {'name':'VSS'}))
        self.pwell = NamedNode(**nodes.get('pwell', {'name':'VPW'}))
        self.nwell = NamedNode(**nodes.get('nwell', {'name':'VNW'}))

        # Logic thresholds
        logic_thresholds = kwargs.get('logic_thresholds', {})
        self.logic_threshold_low = logic_thresholds.get('low', 0.2)
        self.logic_threshold_high = logic_thresholds.get('high', 0.8)
        self.logic_threshold_high_to_low = logic_thresholds.get('high_to_low', 0.5)
        self.logic_threshold_low_to_high = logic_thresholds.get('low_to_high', 0.5)

        # Operating conditions
        self.temperature = kwargs.get('temperature', 25)


class NamedNode:
    """Binds supply node names to voltages"""
    def __init__(self, name, voltage = 0):
        self.name = name
        self.voltage = voltage

    def __str__(self) -> str:
        return f'Name: {self.name}\nVoltage: {self.voltage}'

    def __repr__(self) -> str:
        return f'NamedNode({self.name}, {self.voltage})'



"""Models liberty library groups"""

from charlib.liberty.UnitsSettings import UnitsSettings

# TODO: wire loads, operating conditions, power supplies

class Library:
    """Models a library, which groups cells and their common properties"""

    def __init__(self, name, **attrs):
        """Create a new Library"""
        self.name = name
        self._attrs = attrs
        self.cells = {}

        # General attributes
        self.filename = attrs.get('filename', f'{name}.lib')
        self.technology = 'cmos'
        self.delay_model = 'table_lookup'
        self.bus_naming_style = attrs.get('bus_naming_style', '%s-%d')

        # Delay and Slew attributes
        self.input_threshold_pct_rise = attrs.get('input_threshold_pct_rise', 50)
        self.input_threshold_pct_fall = attrs.get('input_threshold_pct_fall', 50)
        self.output_threshold_pct_rise = attrs.get('output_threshold_pct_rise', 50)
        self.output_threshold_pct_fall = attrs.get('output_threshold_pct_fall', 50)
        self.slew_upper_threshold_pct_rise = attrs.get('slew_upper_threshold_pct_rise', 80)
        self.slew_lower_threshold_pct_rise = attrs.get('slew_lower_threshold_pct_rise', 20)
        self.slew_upper_threshold_pct_fall = attrs.get('slew_upper_threshold_pct_fall', 80)
        self.slew_lower_threshold_pct_fall = attrs.get('slew_lower_threshold_pct_fall', 20)

        # Units
        units = attrs.get('units', {})
        unit_keys = [
            'time',
            'voltage',
            'current',
            'pulling_resistance',
            'capacitive_load',
            'leakage_power',
            'energy',
        ]
        for unit_key in unit_keys:
            if f'{unit_key}_unit' in attrs:
                units[unit_key] = attrs[f'{unit_key}_unit']
        self.units = UnitsSettings(**units)

        # Nominal operating conditions
        self.nom_process = attrs.get('nom_process', 1)
        self.nom_voltage = attrs.get('nom_voltage', 3.3)
        self.nom_temperature = attrs.get('nom_temperature', 25)

    @property
    def time_unit(self):
        """Return self.units.time"""
        return self.units.time

    @property
    def voltage_unit(self):
        """Returns self.units.voltage"""
        return self.units.voltage

    @property
    def current_unit(self):
        """Returns self.units.current"""
        return self.units.current

    @property
    def pulling_resistance_unit(self):
        """Returns self.units.resistance"""
        return self.units.resistance

    @property
    def capacitive_load_unit(self):
        """Returns self.units.capacitance"""
        return self.units.capacitance

    @property
    def leakage_power_unit(self):
        """Returns self.units.power"""
        return self.units.power

    def __getitem__(self, key: str):
        """Returns self[key]. Searches cells by name."""
        return self.cells[key]

    @property
    def attributes(self) -> dict:
        """Return extra attributes on this library"""

    def add_attribute(self, key: str, value: str):
        """Add a new liberty attribute to this cell"""
        self._attrs[key] = str(value)

    def add_cell(self, cell) -> None:
        """Add a cell to the library"""
        self.cells[cell.name] = cell

    def templates(self) -> list:
        templates = []
        for cell in self.cells.values():
            templates.extend(cell.templates())
        return list(set(templates))

    def __str__(self) -> str:
        """Return str(self)"""
        spice_unit = lambda unit : unit.prefixed_unit.str_spice()
        lib_str = [
            f'library ({self.name}) {{',
            f'  technology : {self.technology};',
            f'  delay_model : {self.delay_model};',
            f'  bus_naming_style : "{self.bus_naming_style}";',
            '\n  /* Units */',
            f'  time_unit : "1{spice_unit(self.units.time)}";',
            f'  voltage_unit : "1{spice_unit(self.units.voltage)}";',
            f'  current_unit : "1{spice_unit(self.units.current)}";',
            f'  pulling_resistance_unit : "1{spice_unit(self.units.resistance)}";',
            f'  leakage_power_unit : "1{spice_unit(self.units.power)}";',
            f'  capacitive_load_unit : "1{spice_unit(self.units.capacitance)}";',
            '\n  /* Slew characteristics */',
            f'  slew_upper_threshold_pct_rise : {self.slew_upper_threshold_pct_rise};',
            f'  slew_lower_threshold_pct_rise : {self.slew_lower_threshold_pct_rise};',
            f'  slew_upper_threshold_pct_fall : {self.slew_upper_threshold_pct_fall};',
            f'  slew_lower_threshold_pct_fall : {self.slew_lower_threshold_pct_fall};',
            f'  input_threshold_pct_rise : {self.input_threshold_pct_rise};',
            f'  input_threshold_pct_fall : {self.input_threshold_pct_fall};',
            f'  output_threshold_pct_rise : {self.output_threshold_pct_rise};',
            f'  output_threshold_pct_fall : {self.output_threshold_pct_fall};',
            '\n  /* Operating Conditions */',
            f'  nom_process : {self.nom_process};',
            f'  nom_voltage : {self.nom_voltage};',
            f'  nom_temperature : {self.nom_temperature};',
        ]
        # TODO: Display wire loads, operating conditions, and power supplies

        # Display templates from cells
        lib_str.append('\n  /* Table Templates */')
        for template in self.templates():
            for line in template.split('\n'):
                lib_str.append(f'  {line}')

        # Display cells
        for name, cell in self.cells.items():
            lib_str.append(f'\n  /* {name} */')
            for line in str(cell).split('\n'):
                lib_str.append(f'  {line}')
        
        lib_str.append('}')
        return '\n'.join(lib_str)



from PySpice import Unit

class UnitsSettings:
    def __init__(self, **kwargs) -> None:
        # Initialize using setters
        self.time = kwargs.get('time', 'ns')
        self.voltage = kwargs.get('voltage', 'V')
        self.current = kwargs.get('current', 'uA')
        self.resistance = kwargs.get('pulling_resistance', 'Ω')
        self.capacitance = kwargs.get('capacitive_load', 'pF')
        self.power = kwargs.get('leakage_power', 'nW')
        self.energy = kwargs.get('energy', 'fJ')
        # TODO: Add temperature

    def __str__(self) -> str:
        lines = []
        lines.append(f'Voltage unit:     {str(self.voltage)}')
        lines.append(f'Current unit:     {str(self.current)}')
        lines.append(f'Resistance unit:  {str(self.resistance)}')
        lines.append(f'Capacitance unit: {str(self.capacitance)}')
        lines.append(f'Time unit:        {str(self.time)}')
        lines.append(f'Energy unit:      {str(self.energy)}')
        lines.append(f'Power unit:       {str(self.power)}')
        return '\n'.join(lines)

    @property
    def voltage(self):
        return self._voltage

    @voltage.setter
    def voltage(self, value):
        # Valid symbols are "V" or "Volts"
        if value.lower().endswith('v'):
            prefix_str = value.removesuffix(value[-1])
        elif value.lower().endswith('volts'):
            prefix_str = value.removesuffix(value[-5:])
        else:
            raise ValueError(f'Invalid voltage unit: {value}')
        self._voltage = self._parse_unit(prefix_str, Unit.u_V)

    @property
    def capacitance(self):
        return self._capacitance

    @capacitance.setter
    def capacitance(self, value: str):
        # Valid symbols are "F" or "Farads"
        if value.lower().endswith('f'):
            prefix_str = value.removesuffix(value[-1])
        elif value.lower().endswith('farads'):
            prefix_str = value.removesuffix(value[-6:])
        else:
            raise ValueError(f'Invalid capacitance unit: {value}')
        self._capacitance = self._parse_unit(prefix_str, Unit.u_F)

    @property
    def resistance(self):
        return self._resistance

    @resistance.setter
    def resistance(self, value: str):
        # Valid symbols are "Ω" or "Ohms"
        if value.endswith('Ω'):
            prefix_str = value.removesuffix('Ω')
        elif value.lower().endswith('ohm'):
            prefix_str = value.removesuffix(value[-3:])
        elif value.lower().endswith('ohms'):
            prefix_str = value.removesuffix(value[-4:])
        else:
            raise ValueError(f'Invalid resistance unit: {value}')
        self._resistance = self._parse_unit(prefix_str, Unit.u_Ω)

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, value: str):
        # Valid symbols are "A" or "Amps"
        if value.lower().endswith('a'):
            prefix_str = value.removesuffix(value[-1])
        elif value.lower().endswith('amp'):
            prefix_str = value.removesuffix(value[-3:])
        elif value.lower().endswith('amps'):
            prefix_str = value.removesuffix(value[-4:])
        else:
            raise ValueError(f'Invalid current unit: {value}')
        self._current = self._parse_unit(prefix_str, Unit.u_A)

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value: str):
        # Valid symbols are "s" or "seconds"
        if value.lower().endswith('seconds'):
            prefix_str = value.removesuffix(value[-7:])
        elif value.lower().endswith('s'):
            prefix_str = value.removesuffix(value[-1])
        else:
            raise ValueError(f'Invalid time unit: {value}')
        self._time = self._parse_unit(prefix_str, Unit.u_s)

    @property
    def power(self):
        return self._power

    @power.setter
    def power(self, value: str):
        # Valid symbols are "W" or "Watts"
        if value.lower().endswith('w'):
            prefix_str = value.removesuffix(value[-1])
        elif value.lower().endswith('watts'):
            prefix_str = value.removesuffix(value[-5:])
        else:
            raise ValueError(f'Invalid leakage power unit: {value}')
        self._power = self._parse_unit(prefix_str, Unit.u_W)

    @property
    def energy(self):
        return self._energy
    
    @energy.setter
    def energy(self, value: str):
        # Valid symbols are "J" or "Joules"
        if value.lower().endswith('j'):
            prefix_str = value.removesuffix(value[-1])
        elif value.lower().endswith('joules'):
            prefix_str = value.removesuffix(value[-6:])
        else:
            raise ValueError(f'Invalid energy unit: {value}')
        self._energy = self._parse_unit(prefix_str, Unit.u_J)

    def _parse_unit(self, prefix_str, unit: callable):
        """Convert a metric prefix to its associated exponent"""
        lookup_table = [
            # Prefix             callable
            (('yocto', 'y'),     -24),
            (('zepto', 'z'),     -21),
            (('atto', 'a'),      -18),
            (('femto', 'f'),     -15),
            (('pico', 'p'),      -12),
            (('nano', 'n'),       -9),
            (('micro', 'u', 'μ'), -6),
            (('milli', 'm'),      -3),
            ((''),                 0),
            (('kilo', 'k'),        3),
            (('mega', 'M'),        6),
            (('giga', 'G'),        9),
            (('tera', 'T'),       12),
            (('peta', 'P'),       15),
            (('exa', 'E'),        18),
            (('zetta', 'Z'),      21),
            (('yotta', 'Y'),      24)
        ]
        prefix_str = prefix_str.lower() if len(prefix_str) > 1 else prefix_str # Allow case-insensitive long prefixes
        for (prefixes, exponent) in lookup_table:
            if prefix_str in prefixes: return unit(10.0**exponent).canonise()
        raise ValueError(f'"{prefix_str}" is not a recognized metric prefix! Supported values are: {[prefixes for (prefixes,_) in lookup_table]}')



from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
from PySpice.Unit import *

from charlib.liberty.cell import Pin

@dataclass
class PinTestBinding:
    """Associates a pin to test data, such as state"""
    pin: Pin
    state: str = '0'

    @property
    def direction(self) -> str:
        """Return pin state change direction (if applicable)"""
        if self.state.startswith(('01', 'z1')):
            return 'rise'
        elif self.state.startswith(('10', 'z0')):
            return 'fall'
        else:
            return None

    def __str__(self) -> str:
        return f'{self.pin.name}{self.state}'


class Harness:
    """Characterization parameters for one path through a cell.

    The primary purpose of a harness is to encode mapping of pins to
    states. It also provides some cell-related calculations."""

    def __init__(self, test_manager, pin_state_map: dict) -> None:
        """Create a new Harness"""

        # Parse pin state mapping and set up PinTestBindings
        self._stable_in_ports = []
        self._nontarget_ports = []
        for pin in test_manager.cell.pins.values():
            if pin.name in pin_state_map:
                # Add to targeted or stable ports
                state = pin_state_map[pin.name]
                if state == 'ignore':
                    continue
                binding = PinTestBinding(pin, state)
                if len(state) > 1:
                    if pin.direction == 'input':
                        self._target_in_port = binding
                    elif pin.direction == 'output':
                        self._target_out_port = binding
                else:
                    if pin.direction == 'input':
                        self._stable_in_ports.append(binding)
            else:
                # Add to nontargeted ports
                self._nontarget_ports.append(PinTestBinding(pin))

        # Initialize results from test input slopes and loads
        self.results = {}
        for slew in test_manager.in_slews:
            self.results[str(slew)] = {}
            for load in test_manager.out_loads:
                self.results[str(slew)][str(load)] = {}

    def __str__(self) -> str:
        """Return str(self)"""
        lines = [f'Arc Under Test: {self.arc_str()}']
        if self.stable_in_ports:
            lines.append('    Stable Input Ports:')
            for in_port in self.stable_in_ports:
                lines.append(f'        {in_port.pin.name}: {in_port.state}')
        if self.nontarget_ports:
            lines.append('    Nontarget Ports:')
            for out_port in self.nontarget_ports:
                lines.append(f'        {out_port.pin.name}: {out_port.state}')
        # TODO: Display results if available
        return '\n'.join(lines)

    def short_str(self):
        """Create an abbreviated string for the test vector represented by this harness"""
        harness_str = f'{self.target_in_port.pin.name}={self.target_in_port.state}'
        for in_port in self.stable_in_ports:
            harness_str += f' {in_port.pin.name}={in_port.state}'
        harness_str += f' {self.target_out_port.pin.name}={self.target_out_port.state}'
        for out_port in self.nontarget_ports:
            harness_str += f' {out_port.pin.name}={out_port.state}'
        return harness_str

    def arc_str(self):
        """Return a string representing the test arc"""
        return f'{self.target_in_port.pin.name} ({self.in_direction}) to {self.target_out_port.pin.name} ({self.out_direction})'

    @property
    def target_in_port(self) -> str:
        """Return target input port"""
        return self._target_in_port

    @property
    def stable_in_ports(self) -> list:
        """Return list of stable input ports"""
        return self._stable_in_ports

    @property
    def target_out_port(self) -> str:
        """Return target output ports"""
        return self._target_out_port

    @property
    def nontarget_ports(self) -> list:
        """Return list of nontarget ports"""
        return self._nontarget_ports

    @property
    def in_direction(self) -> str:
        """Return target_in_port.direction"""
        return self.target_in_port.direction

    @property
    def out_direction(self) -> str:
        """Return target_out_port.direction"""
        return self.target_out_port.direction

    @property
    def debug_path(self) -> str:
        """Return folder names for debug results related to this harness."""
        arc_dir = f'{self.target_in_port}_to_{self.target_out_port}'
        stable_dir = '_'.join([str(pin) for pin in self.stable_in_ports])
        return f'{arc_dir}/{stable_dir}'

    def average_propagation_delay(self):
        """Calculates the average propagation delay over all trials"""
        # TODO: Usually we want longest prop delay instead of average
        total_delay = 0.0 @ u_s
        n = 0
        for slope in self.results.keys():
            for load in self.results[slope].keys():
                total_delay += self.results[slope][load]['prop_in_out'] @ u_s
                n += 1
        return total_delay / n

    def _calc_internal_energy(self, slew: str, load: str, energy_meas_high_threshold_voltage: float):
        """Calculates internal energy for a particular slope/load combination"""
        # Fetch calculation parameters, using units to validate calculation
        slew = str(slew)
        load = str(load)
        t_start = self.results[slew][load]['t_energy_start'] @ u_s
        t_end = self.results[slew][load]['t_energy_end'] @ u_s
        q_vdd_dyn = self.results[slew][load]['q_vdd_dyn'] @ u_C
        q_vss_dyn = self.results[slew][load]['q_vss_dyn'] @ u_C
        i_vdd_leak = abs(self.results[slew][load]['i_vdd_leak']) @ u_A
        i_vss_leak = abs(self.results[slew][load]['i_vss_leak']) @ u_A
        # Perform the calculation
        time_delta = (t_end - t_start)
        avg_current = ((i_vdd_leak + i_vss_leak) / 2)
        internal_charge = min(abs(q_vss_dyn), abs(q_vdd_dyn)) - time_delta * avg_current
        return internal_charge * (energy_meas_high_threshold_voltage @ u_V)


class CombinationalHarness (Harness):
    """A CombinationalHarness captures configuration for testing a CombinationalCell."""

    def __init__(self, target_cell, test_vector) -> None:
        """Create a new CombinationalHarness."""
        super().__init__(target_cell, test_vector)
        # Error if we don't have a target input port
        if not self._target_in_port:
            raise ValueError(f'Unable to parse target input port from test vector {test_vector}')

    def plot_energy(self, settings, slews, loads, cell_name):
        """Plot energy vs slew rate vs fanout"""
        # TODO: Consider moving this to Pin, as all the data is eventually stored there anyways
        figure = plt.figure()
        figure.suptitle(f'Cell {cell_name} | Arc: {self.arc_str()}')

        ax = figure.add_subplot(projection='3d')
        ax.set_proj_type('ortho')

        energy_data = []
        for slew in slews:
            energy_row = []
            for load in loads:
                energy = self._calc_internal_energy(slew, load, settings.energy_meas_high_threshold_voltage())
                energy_row.append(float(energy.convert(settings.units.energy.prefixed_unit).value))
            energy_data.append(energy_row)

        # Expand x and y vectors to 2d arrays
        x_data = np.repeat(np.expand_dims(slews, 1), len(loads), 1)
        y_data = np.swapaxes(np.repeat(np.expand_dims(loads, 1), len(slews), 1), 0, 1)

        # Plot energy data
        ax.plot_surface(x_data, y_data, np.asarray(energy_data), cmap='viridis', label='Energy')
        ax.set(xlabel=f'Slew Rate [{str(settings.units.time.prefixed_unit)}]',
               ylabel=f'Fanout [{str(settings.units.capacitance.prefixed_unit)}]',
               zlabel=f'Energy [{str(settings.units.energy.prefixed_unit)}]',
               title='Energy vs. Slew Rate vs. Fanout')


class SequentialHarness (Harness):
    def __init__(self, test_manager, pin_state_map: dict) -> None:
        # Parse internal storage states, clock, set, and reset out of pin mapping
        # Note that set and reset are optional, but must be provided if present
        # on the target cell
        self.set = None
        self.reset = None
        self.clock = PinTestBinding(test_manager.clock, pin_state_map[test_manager.clock.name])
        pin_state_map[test_manager.clock.name] = 'ignore'
        # Set up Reset
        if test_manager.reset:
            self.reset = PinTestBinding(test_manager.reset, pin_state_map[test_manager.reset.name])
            pin_state_map[test_manager.reset.name] = 'ignore'
        # Set up Set
        if test_manager.set:
            self.set = PinTestBinding(test_manager.set, pin_state_map[test_manager.set.name])
            pin_state_map[test_manager.set.name] = 'ignore'
        # TODO: handle flop internal states
        self.flops = []
        super().__init__(test_manager, pin_state_map)

    def short_str(self):
        harness_str = f'{self.clock.pin.name}={self.clock.state} {super().short_str()}'
        if self.set:
            harness_str += f' {self.set.pin.name}={self.set.state}'
        if self.reset:
            harness_str += f' {self.reset.pin.name}={self.reset.state}'
        return harness_str

    @property
    def set_direction(self) -> str:
        if not self.set:
            return None
        return self.set.direction

    @property
    def reset_direction(self) -> str:
        if not self.reset:
            return None
        return self.reset.direction
    
    def invert_set_reset(self):
        self.set.state = self.set.state[::-1] if self.set.state else None
        self.reset.state = self.reset.state[::-1] if self.reset.state else None

    @property
    def timing_sense_constraint(self) -> str:
        # TODO: Check that this is correct
        return f'{self.in_direction}_constraint'

    def _timing_type_with_mode(self, mode) -> str:
        # Determine from target input and direction
        if self.set_direction or self.reset_direction:
            # We're targeting set or reset
            if mode == 'recovery':
                if self.in_direction == 'rise':
                    return f'{mode}_rising'
                else:
                    return f'{mode}_falling'
            elif mode == 'removal':
                if self.in_direction == 'rise':
                    return f'{mode}_falling'
                else:
                    return f'{mode}_rising'
            else:
                return None
        elif not self.target_in_port.pin.name in [*self.flops]:
            # We're targeting an input port
            if mode == 'clock':
                if self.clock.state == '0101':
                    return 'rising_edge'
                else:
                    return 'falling_edge'
            elif mode in ['hold', 'setup']:
                if self.in_direction == 'rise':
                    return f'{mode}_rising'
                else:
                    return f'{mode}_falling'
        # If we get here, most likely the harness isn't configured correctly
        raise ValueError(f'Unable to determine timing type for mode "{mode}"')

    @property
    def timing_type_hold(self) -> str:
        return self._timing_type_with_mode('hold')

    @property
    def timing_type_setup(self) -> str:
        return self._timing_type_with_mode('setup')
    
    @property
    def timing_type_recovery(self) -> str:
        return self._timing_type_with_mode('recovery')
    
    @property
    def timing_type_removal(self) -> str:
        return self._timing_type_with_mode('removal')
    
    @property
    def timing_type_clock(self) -> str:
        return self._timing_type_with_mode('clock')
    
    @property
    def timing_when(self) -> str:
        if self.in_direction == 'rise':
            return self.target_in_port
        else:
            return f'!{self.target_in_port}'

    def plot_energy(self, settings, slews, loads, cell_name):
        pass

# Utilities for working with Harnesses
def filter_harnesses_by_ports(harness_list: list, in_port, out_port) -> list:
    """Finds harnesses in harness_list which target in_port and out_port"""
    return [harness for harness in harness_list 
            if harness.target_in_port.pin == in_port
            and harness.target_out_port.pin == out_port]

def find_harness_by_arc(harness_list: list, in_port, out_port, out_direction) -> Harness:
    harnesses = [harness for harness in filter_harnesses_by_ports(harness_list, in_port, out_port) if harness.out_direction == out_direction]
    if len(harnesses) > 1:
        raise LookupError('Multiple harnesses present in harness_list with the specified arc!')
    elif len(harnesses) < 1:
        raise LookupError('No harnesses present in harness_list with the specified arc!')
    return harnesses[0]

def check_timing_sense(harness_list: list):
    """Checks that all CombinationalHarnesses in harness_list have the same unateness."""
    for harness in harness_list:
        if not harness.timing_sense == harness_list[0].timing_sense:
            return "non_unate"
    return harness_list[0].timing_sense



"""This module contains test managers for various types of standard cells"""

from itertools import product
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Tools.StringTools import str_spice
from PySpice.Unit import *

from charlib.characterizer.functions import Function, registered_functions
from charlib.characterizer.Harness import CombinationalHarness, SequentialHarness, filter_harnesses_by_ports, find_harness_by_arc
from charlib.characterizer.LogicParser import parse_logic
from charlib.liberty.cell import Cell, Pin, TimingData, TableTemplate

class TestManager:
    """A test manager for a standard cell"""
    def __init__ (self, name: str, in_ports: str|list, out_ports: list|None, functions: str|list, **kwargs):
        """Create a new TestManager for a standard cell.

        A TestManager manages cell data, runs simulations on cells, and stores results
        on the cell.
        
        :param name: cell name
        :param in_ports: a list of input pin names
        :param out_ports: a list of output pin names
        :param functions: a list of functions implemented by each of the cell's outputs (in verilog syntax)
        :param **kwargs: a dict of configuration and test parameters for the cell, including
            - netlist: path to the cell's spice netlist
            - model: path to transistor spice models
            - slews: input slew rates to test
            - loads: output capacitave loads to test
            - simulation_timestep: the time increment to use during simulation"""
        # Initialize the cell under test
        self._cell = Cell(name, kwargs.get('area', 0))
        for pin_name in in_ports:
            self.cell.add_pin(pin_name, 'input')
        for pin_name in out_ports:
            self.cell.add_pin(pin_name, 'output')

        # Parse functions and add to pins
        if isinstance(functions, str):
            functions = functions.upper().split() # Capitalize and split on space, then proceed
        if isinstance(functions, list):
            # Should be in the format ['Y=expr1', 'Z=expr2']
            for func in functions:
                if '=' in func:
                    func_pin, expr = func.split('=')
                    # Check for nonblocking assign in LHS of equation
                    if '<' in func_pin:
                        func_pin = func_pin.replace('<','').strip()
                    for pin_name in out_ports:
                        if pin_name == func_pin:
                            if parse_logic(expr):
                                # Check if we already recognize this function
                                function = Function(expr)
                                for reg_name, reg_func in registered_functions.items():
                                    if reg_func == function:
                                        # Copy test vectors
                                        print(f'Recognized {self.cell.name} pin {pin_name} function as {reg_name}')
                                        function.stored_test_vectors = reg_func.test_vectors
                                self.cell[pin_name].function = function
                            else:
                                raise ValueError(f'Invalid function "{expr}"')
                else:
                    raise ValueError(f'Expected an expression of the form "Y=A Z=B" for cell function, got "{func}"')

        # Characterization settings
        self.netlist = kwargs.get('netlist')
        self.models = kwargs.get('models', [])
        self._in_slews = kwargs.get('slews', [])
        self._out_loads = kwargs.get('loads', [])
        self.sim_timestep = kwargs.get('simulation_timestep', min(self.in_slews)/4.0)

        # Behavioral/internal-use settings
        self.plots = kwargs.get('plots', [])
        self._is_exported = False

    @property
    def cell(self) -> Cell:
        """Return the cell under test"""
        return self._cell

    @property
    def in_ports(self) -> list:
        """Return cell input io pins."""
        return [pin for pin in self.cell.pins.values() if pin.direction == 'input' and pin.is_io()]

    @property
    def out_ports(self) -> list:
        """Return cell output io pins."""
        return [pin for pin in self.cell.pins.values() if pin.direction == 'output' and pin.is_io()]

    @property
    def functions(self) -> list:
        """Return a list of functions on this cell's output pins."""
        return [pin.function for pin in self.out_ports]

    @property
    def models(self) -> list:
        """Return cell models"""
        return self._models

    @models.setter
    def models(self, value):
        models = []
        """Set paths to cell transistor models"""
        for model in value:
            modelargs = model.split()
            path = Path(modelargs.pop(0))
            if modelargs: # If the list is not empty (e.g. there is a section parameter)
                section = modelargs.pop(0)
                # Use a tuple so that this is included with .lib path section syntax
                if not path.is_file():
                    raise ValueError(f'Invalid model {path} {section}: {path} is not a file')
                models.append((path, section))
            else:
                if path.is_dir():
                    models.append(SpiceLibrary(path))
                elif path.is_file():
                    models.append(path)
                else:
                    raise FileNotFoundError(f'File {value} not found')
        self._models = models

    def _include_models(self, circuit):
        """Include models in the circuit netlist."""
        for model in self.models:
            if isinstance(model, SpiceLibrary):
                for device in self.used_models():
                    # TODO: Handle the case where we have multiple spice libraries
                    circuit.include(model[device])
            elif isinstance(model, Path):
                circuit.include(model)
            elif isinstance(model, tuple):
                circuit.lib(*model)

    @property
    def netlist(self) -> str:
        """Return path to cell netlist."""
        return self._netlist

    @netlist.setter
    def netlist(self, value):
        """Set path to cell netlist."""
        if isinstance(value, (str, Path)):
            if not Path(value).is_file():
                raise ValueError(f'Invalid value for netlist: {value} is not a file')
            self._netlist = Path(value)
        else:
            raise TypeError(f'Invalid type for netlist: {type(value)}')

    def definition(self) -> str:
        """Return the cell's spice definition"""
        # Search the netlist file for the circuit definition
        with open(self.netlist, 'r') as file:
            for line in file:
                if self.cell.name in line.upper() and 'SUBCKT' in line.upper():
                    file.close()
                    return line
            # If we reach this line before returning, the netlist file doesn't contain a circuit definition
            file.close()
            raise ValueError(f'No cell definition found in netlist {self.netlist}')

    def instance(self) -> str:
        """Return a subcircuit instantiation for this cell."""
        # Reorganize the definition into an instantiation with instance name XDUT
        # TODO: Instance name should probably be configurable from library settings
        instance = self.definition().split()[1:]  # Delete .subckt
        instance.append(instance.pop(0))        # Move circuit name to last element
        instance.insert(0, 'XDUT')                # Insert instance name
        return ' '.join(instance)

    def used_models(self) -> list:
        """Return a list of subcircuits used by this cell."""
        subckts = []
        with open(self.netlist, 'r') as file:
            for line in file:
                if line.lower().startswith('x'):
                    # Get the subckt name
                    # This should be the last item that doesn't contain =
                    for term in reversed(line.split()):
                        if '=' not in term:
                            subckts.append(term)
                            break
            file.close()
        return subckts

    @property
    def in_slews(self) -> list:
        """Return slew rates to use during testing."""
        return self._in_slews

    def add_in_slew(self, value: float):
        """Add a slew rate to the list of slew rates."""
        self._in_slews.append(float(value))

    @property
    def out_loads(self) -> list:
        """Return output capacitive loads to use during testing."""
        return self._out_loads

    def add_out_load(self, value: float):
        """Add a load to the list of output loads."""
        self._out_loads.append(float(value))

    @property
    def plots(self) -> list:
        """Return plotting configuration."""
        return self._plots
    
    @plots.setter
    def plots(self, value):
        """Set plot configuration

        :param value: a str or list specifying which plot types to generate."""
        if value == 'all':
            self._plots = ['io', 'delay', 'power']
        elif value == 'none':
            self._plots = []
        elif isinstance(value, list):
            self._plots = value
        else:
            raise ValueError(f'Invalid value for plots: "{value}"')

    @property
    def is_exported(self) -> bool:
        """Return whether the results have been exported"""
        return self._is_exported

    def set_exported(self):
        """Set a flag that this test manager's results have been exported"""
        self._is_exported = True

    def _run_input_capacitance(self, settings, target_pin):
        """Measure the input capacitance of target_pin.

        Assuming a black-box model, treat the cell as a grounded capacitor with fixed capacitance.
        Perform an AC sweep on the circuit and evaluate the capacitance as d/ds(i(s)/v(s))."""
        print(f'Running input_capacitance for pin {target_pin} of cell {self.cell.name}')
        vdd = settings.vdd.voltage * settings.units.voltage
        vss = settings.vss.voltage * settings.units.voltage
        # TODO: Make these values configurable from settings
        f_start = 10 @ u_Hz
        f_stop = 10 @ u_GHz
        r_in = 10 @ u_GOhm
        i_in = 1 @ u_uA
        r_out = 10 @ u_GOhm
        c_out = 1 @ u_pF

        # Initialize circuit
        circuit = Circuit(f'{self.cell.name}_pin_{target_pin}_cap')
        self._include_models(circuit)
        circuit.include(self.netlist)
        circuit.V('dd', 'vdd', circuit.gnd, vdd)
        circuit.V('ss', 'vss', circuit.gnd, vss)
        circuit.I('in', circuit.gnd, 'vin', f'DC 0 AC {str_spice(i_in)}')
        circuit.R('in', circuit.gnd, 'vin', r_in)

        # Initialize device under test and wire up ports
        ports = self.definition().upper().split()[1:]
        subcircuit_name = ports.pop(0)
        connections = []
        for port in ports:
            if port == target_pin:
                connections.append('vin')
            elif port == settings.vdd.name.upper():
                connections.append('vdd')
            elif port == settings.vss.name.upper():
                connections.append('vss')
            else:
                # Add a resistor and capacitor to each output
                circuit.C(port, f'v{port}', circuit.gnd, c_out)
                circuit.R(port, f'v{port}', circuit.gnd, r_out)
                connections.append(f'v{port}')
        circuit.X('dut', subcircuit_name, *connections)
        simulator = circuit.simulator(temperature=settings.temperature,
                                      nominal_temperature=settings.temperature,
                                      simulator=settings.simulator)

        # Log simulation
        # Path should be debug_dir/cell_name/input_capacitance/pin
        if settings.debug:
            debug_path = settings.debug_dir / self.cell.name / 'input_capacitance'
            debug_path.mkdir(parents=True, exist_ok=True)
            with open(debug_path/f'{target_pin}.sp', 'w') as spice_file:
                spice_file.write(str(simulator))

        # Measure capacitance as the slope of the conductance
        analysis = simulator.ac('dec', 100, f_start, f_stop)
        impedance = np.abs(analysis.vin)/i_in
        [capacitance, _] = np.polyfit(analysis.frequency, np.reciprocal(impedance)/(2*np.pi), 1)

        return capacitance


class CombinationalTestManager(TestManager):
    """A combinational cell test manager"""

    def characterize(self, settings):
        """Characterize a combinational cell"""
        # Measure input capacitance for all input pins
        for pin in self.in_ports:
            input_capacitance = self._run_input_capacitance(settings, pin.name) @ u_F
            self.cell[pin.name].capacitance = input_capacitance.convert(settings.units.capacitance.prefixed_unit).value

        # Run delay simulation for all test vectors of each function
        for out_port in self.out_ports:
            unsorted_harnesses = []
            # Run characterization
            for test_vector in out_port.function.test_vectors:
                # Map pins to test vector
                inputs = out_port.function.operands
                state_map = dict(zip([*inputs, out_port.name], test_vector))

                # Generate harness
                harness = CombinationalHarness(self, state_map)
                trial_name = f'delay {self.cell.name} {harness.short_str()}'

                # Run delay characterization
                for slew in self.in_slews:
                    for load in self.out_loads:
                        self._run_delay(settings, harness, slew, load, trial_name)
                unsorted_harnesses.append(harness)

            # Filter out harnesses that aren't worst-case conditions
            # We should be left with the critical path rise and fall harnesses for each i/o path
            harnesses = []
            for in_port in self.in_ports:
                for direction in ['rise', 'fall']:
                    # Iterate over harnesses that match output, input, and direction
                    matching_harnesses = [harness for harness in filter_harnesses_by_ports(unsorted_harnesses, in_port, out_port) if harness.out_direction == direction]
                    worst_case_harness = matching_harnesses[0]
                    for harness in matching_harnesses:
                        # FIXME: Currently we compare by average prop delay. Consider alternative strategies
                        if worst_case_harness.average_propagation_delay() < harness.average_propagation_delay():
                            worst_case_harness = harness # This harness is worse
                    harnesses.append(worst_case_harness)

            # Store propagation and transient delay in pin timing tables
            for in_port in self.in_ports:
                delay_timing = TimingData(in_port.name)
                for direction in ['rise', 'fall']:
                    # Identify the correct harness
                    harness = find_harness_by_arc(harnesses, in_port, out_port, direction)

                    # Construct the table
                    index_1 = [str(slew) for slew in self.in_slews]
                    index_2 = [str(load) for load in self.out_loads]
                    prop_values = []
                    tran_values = []
                    for slew in index_1:
                        for load in index_2:
                            result = harness.results[slew][load]
                            prop_value = (result['prop_in_out'] @ u_s).convert(settings.units.time.prefixed_unit).value
                            prop_values.append(f'{prop_value:7f}')
                            tran_value = (result['trans_out'] @ u_s).convert(settings.units.time.prefixed_unit).value
                            tran_values.append(f'{tran_value:7f}')
                    template = TableTemplate()
                    template.name = f'delay_template_{len(index_1)}x{len(index_2)}'
                    template.variables = ['input_net_transition', 'total_output_net_capacitance']
                    delay_timing.add_table(f'cell_{direction}', template, prop_values, index_1, index_2)
                    delay_timing.add_table(f'{direction}_transition', template, tran_values, index_1, index_2)
                self.cell[out_port.name].timings.append(delay_timing)

            # Display plots
            if 'io' in self.plots:
                [self.plot_io(settings, harness) for harness in harnesses]
            if 'delay' in self.plots:
                [self.cell[out_pin.name].plot_delay(settings, self.cell.name) for out_pin in self.out_ports]
            if 'energy' in self.plots:
                print("Energy plotting not yet supported") # TODO: Add correct energy measurement procedure
            if plt.get_figlabels():
                plt.tight_layout()
                plt.show()

        return self.cell

    def _run_delay(self, settings, harness: CombinationalHarness, slew, load, trial_name):
        print(f'Running {trial_name} with slew={slew*settings.units.time}, load={load*settings.units.capacitance}')
        harness.results[str(slew)][str(load)] = self._run_delay_trial(settings, harness, slew, load)

    def _run_delay_trial(self, settings, harness: CombinationalHarness, slew, load):
        """Run delay measurement for a single trial"""
        # Set up parameters
        data_slew = slew * settings.units.time
        t_start = data_slew
        t_end = t_start + data_slew
        t_simend = 10000 * data_slew
        vdd = settings.vdd.voltage * settings.units.voltage
        vss = settings.vss.voltage * settings.units.voltage

        # Initialize circuit
        circuit = Circuit(f'{self.cell.name}_delay')
        self._include_models(circuit)
        circuit.include(self.netlist)
        (v_start, v_end) = (vss, vdd) if harness.in_direction == 'rise' else (vdd, vss)
        pwl_values = [(0, v_start), (t_start, v_start), (t_end, v_end), (t_simend, v_end)]
        circuit.PieceWiseLinearVoltageSource('in', 'vin', circuit.gnd, values=pwl_values)
        circuit.V('high', 'vhigh', circuit.gnd, vdd)
        circuit.V('low', 'vlow', circuit.gnd, vss)
        circuit.V('dd_dyn', 'vdd_dyn', circuit.gnd, vdd)
        circuit.V('ss_dyn', 'vss_dyn', circuit.gnd, vss)
        circuit.V('o_cap', 'vout', 'wout', circuit.gnd)
        circuit.C('0', 'wout', 'vss_dyn', load * settings.units.capacitance)

        # Initialize device under test subcircuit and wire up ports
        ports = self.definition().upper().split()[1:]
        subcircuit_name = ports.pop(0)
        connections = []
        for port in ports:
            if port == harness.target_in_port.pin.name:
                connections.append('vin')
            elif port == harness.target_out_port.pin.name:
                connections.append('vout')
            elif port == settings.vdd.name.upper():
                connections.append('vdd_dyn')
            elif port == settings.vss.name.upper():
                connections.append('vss_dyn')
            elif port in [pin.pin.name for pin in harness.stable_in_ports]:
                for stable_port in harness.stable_in_ports:
                    if port == stable_port.pin.name:
                        if stable_port.state == '1':
                            connections.append('vhigh')
                        elif stable_port.state == '0':
                            connections.append('vlow')
                        else:
                            raise ValueError(f'Invalid state identified during simulation setup for port {port}: {state}')
            else:
                connections.append('wfloat0') # Float any unrecognized ports
        if len(connections) is not len(ports):
            raise ValueError(f'Failed to match all ports identified in definition "{self.definition().strip()}"')
        circuit.X('dut', subcircuit_name, *connections)

        # Initialize simulator
        simulator = circuit.simulator(temperature=settings.temperature,
                                      nominal_temperature=settings.temperature,
                                      simulator=settings.simulator)
        simulator.options('autostop', 'nopage', 'nomod', post=1, ingold=2, trtol=1)

        # Measure delay
        pct_vdd = lambda x : x * settings.vdd.voltage
        match harness.in_direction:
            case 'rise':
                v_prop_start = settings.logic_threshold_low_to_high
            case 'fall':
                v_prop_start = settings.logic_threshold_high_to_low
        match harness.out_direction:
            case 'rise':
                v_prop_end = settings.logic_threshold_low_to_high
                v_trans_start = settings.logic_threshold_low
                v_trans_end = settings.logic_threshold_high
            case 'fall':
                v_prop_end = settings.logic_threshold_high_to_low
                v_trans_start = settings.logic_threshold_high
                v_trans_end = settings.logic_threshold_low
        simulator.measure('tran', 'prop_in_out',
                        f'trig v(vin) val={pct_vdd(v_prop_start)} {harness.in_direction}=1',
                        f'targ v(vout) val={pct_vdd(v_prop_end)} {harness.out_direction}=1')
        simulator.measure('tran', 'trans_out',
                        f'trig v(vout) val={pct_vdd(v_trans_start)} {harness.out_direction}=1',
                        f'targ v(vout) val={pct_vdd(v_trans_end)} {harness.out_direction}=1')

        # Log simulation
        # Path should be debug_dir/cell_name/delay/arc/slew/load/
        if settings.debug:
            debug_path = settings.debug_dir / self.cell.name / 'delay' / harness.debug_path / \
                         f'slew_{slew}' / f'load_{load}'
            debug_path.mkdir(parents=True, exist_ok=True)
            with open(debug_path/'delay.sp', 'w') as spice_file:
                spice_file.write(str(simulator))

        # Run transient analysis
        step_time = min(self.sim_timestep*settings.units.time, t_simend/1000)
        return simulator.transient(step_time=step_time, end_time=t_simend)

    def plot_io(self, settings, harness):
        """Plot I/O voltages vs time"""
        # TODO: Look for ways to generate fewer plots here - maybe a creative 3D plot
        figures = []
        # Group data by slew rate so that inputs are the same
        for slew in self.in_slews:
            # Generate plots for Vin and Vout
            figure, (ax_i, ax_o) = plt.subplots(2,
                sharex=True,
                height_ratios=[3, 7],
                label=f'{self.cell.name} | {harness.arc_str()} | {str(slew*settings.units.time)}'
            )
            volt_units = str(settings.units.voltage.prefixed_unit)
            time_units = str(settings.units.time.prefixed_unit)
            ax_i.set(
                ylabel=f'Vin (pin {harness.target_in_port.pin.name}) [{volt_units}]',
                title='I/O Voltages vs. Time'
            )
            ax_o.set(
                ylabel=f'Vout (pin {harness.target_out_port.pin.name}) [{volt_units}]',
                xlabel=f'Time [{time_units}]'
            )
            for load in self.out_loads:
                analysis = harness.results[str(slew)][str(load)]
                ax_o.plot(analysis.time / settings.units.time, analysis.vout, label=f'Fanout={load*settings.units.capacitance}')
            ax_o.legend()
            ax_i.plot(analysis.time / settings.units.time, analysis.vin)

            # Add lines indicating logic levels and timing
            for ax in [ax_i, ax_o]:
                ax.grid()
                for level in [settings.logic_threshold_low, settings.logic_threshold_high]:
                    ax.axhline(level*settings.vdd.voltage, color='0.5', linestyle='--')
                for t in [slew, 2*slew]:
                    ax.axvline(float(t), color='r', linestyle=':')

            figures.append(figure)
        return figures
        

class SequentialTestManager(TestManager):
    """A sequential cell test manager"""

    def __init__(self, name: str, in_ports: list, out_ports: list, clock: str, flops: str, function: str, **kwargs):
        super().__init__(name, in_ports, out_ports, function, **kwargs)
        # TODO: Use flops in place of functions for sequential cells
        self.set = kwargs.get('set')
        self.reset = kwargs.get('reset')
        self.clock = clock
        self.flops = flops

        self._clock_slew = kwargs.get('clock_slew', 0)

        # Setup and Hold time search parameters
        self.setup_time_range = kwargs.get('setup_time_range', [0.1, 1])
        self.hold_time_range = kwargs.get('hold_time_range', [0.1, 1])

    @property
    def clock(self) -> Pin:
        """Return clock pin"""
        return self.cell[self.clock_name]

    @property
    def clock_name(self) -> str:
        """Return clock pin name."""
        return self._clock_name

    @property
    def clock_trigger(self) -> str:
        """Return clock trigger type."""
        return self._clock_trigger

    @clock.setter
    def clock(self, value: str):
        """Assign clock trigger and pin"""
        (self._clock_trigger, pin) = _parse_triggered_pin(value, 'clock')
        self._clock_name = pin.name
        self.cell.add_pin(pin.name, pin.direction, pin.role)

    @property
    def clock_slew(self) -> float:
        """Return clock slew rate"""
        if self.in_slews and not self._clock_slew:
            return min(self.in_slews)
        return float(self._clock_slew)

    @property
    def set(self):
        """Return set pin"""
        return self.cell.pins.get(self.set_name)

    @property
    def set_name(self) -> str:
        """Return set pin name"""
        return self._set_name

    @property
    def set_trigger(self) -> str:
        "Return set pin trigger type"
        return self._set_trigger

    @set.setter
    def set(self, value):
        """Assign set pin and trigger"""
        if value is not None:
            (self._set_trigger, pin) = _parse_triggered_pin(value, 'set')
            self._set_name = pin.name
            self.cell.add_pin(pin.name, pin.direction, pin.role)
        else:
            self._set_name = None

    @property
    def reset(self):
        """Return reset pin"""
        return self.cell.pins.get(self.reset_name)

    @property
    def reset_name(self) -> str:
        """Return reset pin name"""
        return self._reset_name
    
    @property
    def reset_trigger(self) -> str:
        """Return reset trigger type"""
        return self._reset_trigger

    @reset.setter
    def reset(self, value):
        """Assign reset pin and trigger"""
        if value is not None:
            (self._reset_trigger, pin) = _parse_triggered_pin(value, 'reset')
            self._reset_name = pin.name
            self.cell.add_pin(pin.name, pin.direction, pin.role)
        else:
            self._reset_name = None

    @property
    def flops(self) -> list:
        # TODO: Use flops in place of functions for sequential cells
        return self._flops

    @flops.setter
    def flops(self, value):
        # TODO: Use flops in place of functions for sequential cells
        if isinstance(value, str):
            self._flops = value.split()
        elif isinstance(value, list):
            self._flops = value
        else:
            raise TypeError(f'Invalid type for sequential cell flop names: {type(value)}')

    def characterize(self, settings):
        """Run Delay, Recovery & Removal characterization for a sequential cell"""
        # Measure input capacitance for all input pins
        in_cap_pins = [*self.in_ports, self.clock]
        if self.set:
            in_cap_pins += [self.set]
        if self.reset:
            in_cap_pins += [self.reset]
        for pin in in_cap_pins:
            input_capacitance = self._run_input_capacitance(settings, pin.name) @ u_F
            self.cell[pin.name].capacitance = input_capacitance.convert(settings.units.capacitance.prefixed_unit).value

        # Save test results to cell
        normalize_t_units = lambda value: (value @ u_s).convert(settings.units.time.prefixed_unit).value

        for out_port in self.out_ports:
            unsorted_harnesses = []
            # Generate Harnesses and run characterization
            for test_vector in out_port.function.test_vectors:
                # Map pins
                inputs = out_port.function.operands
                state_map = dict(zip([*inputs, out_port.name], test_vector))
                state_map[self.clock.name] = '0101' if self.clock_trigger == 'posedge' else '1010'
                if self.set:
                    state_map[self.set.name] = '0' if self.set_trigger == 'posedge' else '1'
                if self.reset:
                    state_map[self.reset.name] = '0' if self.reset_trigger == 'posedge' else '1'
                # TODO: Add flops

                # Generate harness
                harness = SequentialHarness(self, state_map)
                trial_name = f'delay {self.cell.name} {harness.short_str()}'

                # Run characterization
                for slew in self.in_slews:
                    for load in self.out_loads:
                        harness.results[str(slew)][str(load)] = self._run_delay(settings, harness, slew, load, trial_name)
                unsorted_harnesses.append(harness)

            # TODO: Filter out harnesses that aren't worst-case conditions
            harnesses = unsorted_harnesses

            # Store timing results
            for in_port in self.in_ports: # TODO: Add set and reset
                index_1 = [str(slew) for slew in self.in_slews]
                index_2 = [str(load) for load in self.out_loads]

                # Set up timing groups and table templates
                clock_edge = 'rising' if self.clock_trigger == 'posedge' else 'falling'
                delay_template = TableTemplate()
                delay_template.name = f'delay_template_{len(index_1)}x{len(index_2)}'
                delay_template.variables = ['input_net_transition', 'total_output_net_capacitance']
                delay_timing = TimingData(out_port.name, f'{clock_edge}_edge')
                setup_template = TableTemplate()
                setup_template.name = f'setup_template_{len(index_1)}x{len(index_2)}'
                setup_template.variables = ['related_pin_transition', 'constrained_pin_transition']
                setup_timing = TimingData(self.clock_name, f'setup_{clock_edge}')
                hold_template = TableTemplate()
                hold_template.name = f'hold_template_{len(index_1)}x{len(index_2)}'
                hold_template.variables = setup_template.variables
                hold_timing = TimingData(self.clock_name, f'hold_{clock_edge}')
                for direction in ['rise', 'fall']:
                    # Fetch and format data for timing tables
                    harness = find_harness_by_arc(harnesses, in_port, out_port, direction)
                    prop_values = []
                    tran_values = []
                    setup_values = []
                    hold_values = []
                    for slew in index_1:
                        for load in index_2:
                            result = harness.results[slew][load]
                            prop_values.append(f'{normalize_t_units(result["prop_in_out"]):7f}')
                            tran_values.append(f'{normalize_t_units(result["trans_out"]):7f}')
                            setup_values.append(f'{normalize_t_units(result["t_setup"]):7f}')
                            hold_values.append(f'{normalize_t_units(result["t_hold"]):7f}')

                    # Store propagation and transient delays on the output pin
                    delay_timing.add_table(f'cell_{direction}', delay_template, prop_values, index_1, index_2)
                    delay_timing.add_table(f'{direction}_transition', delay_template, tran_values, index_1, index_2)

                    # Store setup and hold constraints on the input pin
                    setup_timing.add_table(f'{direction}_constraint', setup_template, setup_values, index_1, index_2)
                    hold_timing.add_table(f'{direction}_constraint', hold_template, hold_values, index_1, index_2)

                # Add timing groups to pins
                self.cell[out_port.name].timings.append(delay_timing)
                self.cell[in_port.name].timings.append(setup_timing)
                self.cell[in_port.name].timings.append(hold_timing)

            # TODO: Store internal power results

            # Display plots
            if 'io' in self.plots:
                [self.plot_io(settings, harness) for harness in harnesses]
            if 'delay' in self.plots:
                [self.cell[out_pin.name].plot_delay(settings, self.cell.name) for out_pin in self.out_ports]
            if 'energy' in self.plots:
                pass # TODO
            if plt.get_figlabels():
                plt.tight_layout()
                plt.show()

        return self.cell

    def _run_delay(self, settings, harness: SequentialHarness, slew, load, trial_name):
        """Run a single sequential delay trial"""
        # Set up slew and load parameters
        t_slew = slew * settings.units.time
        c_load = load * settings.units.capacitance
        debug_path = settings.debug_dir / self.cell.name / 'delay' / harness.debug_path / \
                     f'slew_{slew}' / f'load_{load}'
    
        print(f'Running sequential {trial_name} with slew={str(t_slew)}, load={str(c_load)}')
        t_stab = self._find_stabilizing_time(settings, harness, t_slew, c_load, debug_path)
        (t_setup, t_hold) = self._find_setup_hold_delay(settings, harness, t_slew, c_load, t_stab, debug_path)

        # Characterize using identified setup and hold time
        simulator, timings = self._build_test_circuit('delay', settings, harness, t_slew, c_load, t_setup, t_hold, t_stab)
        return self._measure_cell_delays(settings, harness, simulator, timings, debug_path)

    def _find_stabilizing_time(self, settings, harness, t_slew, c_load, debug_path):
        """Find a reasonable stablilizing time for the current configuration.

        The stabilizing time is the delay between the first half of the procedure, where we zero
        out any initial state, and the second half of the procedure, where we measure delay
        characteristics. It's important to minimize stabilizing time as it has a major effect on
        total simulation time."""
        # Run a single simulation and measure the time it takes for the Q output to change from 1%
        # of vdd to 99% of vdd
        t_stab = 500 * max(self.in_slews) * settings.units.time
        t_setup = max(self.setup_time_range) * settings.units.time
        t_hold = max(self.hold_time_range) * settings.units.time
        sim, t = self._build_test_circuit('stabilizing', settings, harness, t_slew, c_load, t_setup, t_hold, t_stab)

        # Measure time it takes for Q to stabilize
        match harness.out_direction:
            case 'rise':
                v_start = 0.01 * settings.vdd.voltage
                v_end = 0.99 * settings.vdd.voltage
            case 'fall':
                v_start = 0.99 * settings.vdd.voltage
                v_end = 0.01 * settings.vdd.voltage
        sim.measure('tran', 't_stabilizing',
            f'trig v(vout) val={v_start} {harness.out_direction}=1',
            f'targ v(vout) val={v_end} {harness.out_direction}=1')

        # Log simulation
        if settings.debug:
            debug_path.mkdir(parents=True, exist_ok=True)
            with open(debug_path/'stabilizing.sp', 'w') as spice_file:
                spice_file.write(str(sim))
            
        step_time = t['sim_end']/5000 # Run with low precision
        results = sim.transient(step_time=step_time, end_time=t['sim_end'])
        return results['t_stabilizing'] @ u_s

    def _find_setup_hold_delay(self, settings, harness, t_slew, c_load, t_stabilizing, debug_path):
        """Calculate setup and hold time.

        Calculate the minimum setup and hold time for the current configuration, accounting for
        interdependence between the two. Uses the procedure proposed by Salman et. al.; See
        https://ieeexplore.ieee.org/document/4167994"""
        # Identify msp
        th = max(self.hold_time_range) * settings.units.time
        t_setup_min = self._sweep_ts(settings, harness, t_slew, c_load, t_stabilizing, th, 'ts_min', debug_path)
        t_hold_max = self._sweep_th(settings, harness, t_slew, c_load, t_stabilizing, t_setup_min, 'th_max', debug_path)

        # Identify mhp
        ts = max(self.setup_time_range) * settings.units.time
        t_hold_min = self._sweep_th(settings, harness, t_slew, c_load, t_stabilizing, ts, 'th_min', debug_path)
        t_setup_max = self._sweep_ts(settings, harness, t_slew, c_load, t_stabilizing, t_hold_min, 'ts_max', debug_path)

        # Interpolate mshp along the contour formed by msp, mhp
        # For now we use a simple average, which may be overly pessimistic
        # We'll also increase setup by 40% as a sort of "safety factor"
        mshp = (1.4*(t_setup_min+t_setup_max)/2, (t_hold_min+t_hold_max)/2)
        return mshp

    def _sweep_ts(self, settings, harness, t_slew, c_load, t_stabilizing, t_hold, title, debug_path):
        """Perform a binary search to find the minimum viable t_setup with the given t_hold"""
        t_step = self.sim_timestep * settings.units.time
        ts_max = max(self.setup_time_range) * settings.units.time
        ts_min = min(self.setup_time_range) * settings.units.time
        ts = ts_max
        i = 0
        while ts - ts_min > t_step:
            i += 1
            sim, t = self._build_test_circuit(f'{title}_{i}', settings, harness, t_slew, c_load, ts, t_hold, t_stabilizing)
            try:
                self._measure_c2q(settings, harness, sim, t, debug_path)
            except NameError:
                ts_min = ts
                ts = (ts_max + ts_min) / 2
                continue
            ts_max = ts
            ts = (ts_max + ts_min) / 2
        return ts_max

    def _sweep_th(self, settings, harness, t_slew, c_load, t_stabilizing, t_setup, title, debug_path):
        """Perform a binary search to find the minimum viable t_hold with the given t_setup"""
        t_step = self.sim_timestep * settings.units.time
        th_max = max(self.hold_time_range) * settings.units.time
        th_min = min(self.hold_time_range) * settings.units.time
        th = th_max
        i = 0
        while th - th_min > t_step:
            i += 1
            sim, t = self._build_test_circuit(f'{title}_{i}', settings, harness, t_slew, c_load, t_setup, th, t_stabilizing)
            try:
                self._measure_c2q(settings, harness, sim, t, debug_path)
            except NameError:
                th_min = th
                th = (th_max + th_min) / 2
                continue
            th_max = th
            th = (th_max + th_min) / 2
        return th_max

    def _build_test_circuit(self, title, settings, harness, t_slew, c_load, t_setup, t_hold, t_stabilizing):
        """Construct the circuit simulator object with the provided test parameters"""
        # Set up parameters
        clk_slew = self.clock_slew * settings.units.time
        vdd = settings.vdd.voltage * settings.units.voltage
        vss = settings.vss.voltage * settings.units.voltage

        # Set up timing parameters for clock and data events
        t = {}
        t['clk_edge_1_start'] = t_setup
        t['clk_edge_1_end'] = t['clk_edge_1_start'] + clk_slew
        t['clk_edge_2_start'] = t['clk_edge_1_end'] + max(t_setup, t_hold)
        t['clk_edge_2_end'] = t['clk_edge_2_start'] + clk_slew
        t['removal'] = t['clk_edge_2_end'] + t_hold # initial state has now been zeroed out
        t['data_edge_1_start'] = t['removal'] + t_stabilizing # wait for the system to stabilize
        t['data_edge_1_end'] = t['data_edge_1_start'] + t_slew
        t['clk_edge_3_start'] = t['data_edge_1_start'] + t_slew/2 + t_setup + clk_slew/2
        t['clk_edge_3_end'] = t['clk_edge_3_start'] + clk_slew
        t['data_edge_2_start'] = t['clk_edge_3_start'] + clk_slew/2 + t_hold + t_slew/2
        t['data_edge_2_end'] = t['data_edge_2_start'] + t_slew
        t['sim_end'] = t['data_edge_2_end'] + 2*t_stabilizing # wait for the system to stabilize

        # Initialize circuit
        circuit = Circuit(title)
        self._include_models(circuit)
        circuit.include(self.netlist)
        circuit.V('high', 'vhigh', circuit.gnd, vdd)
        circuit.V('low', 'vlow', circuit.gnd, vss)
        circuit.V('dd_dyn', 'vdd_dyn', circuit.gnd, vdd)
        circuit.V('ss_dyn', 'vss_dyn', circuit.gnd, vss)
        circuit.V('o_cap', 'vout', 'wout', 0)
        circuit.C('0', 'wout', 'vss_dyn', c_load)

        # Set up clock input
        (v0, v1) = (vdd, vss) if harness.timing_type_clock == 'falling_edge' else (vss, vdd)
        circuit.PieceWiseLinearVoltageSource('cin', 'vcin', circuit.gnd, values=[
            (0, v0),
            (t['clk_edge_1_start'], v0),
            (t['clk_edge_1_end'], v1),
            (t['clk_edge_2_start'], v1),
            (t['clk_edge_2_end'], v0),
            (t['clk_edge_3_start'], v0),
            (t['clk_edge_3_end'], v1),
            (t['sim_end'], v1)
        ])

        # Set up data input node
        (v0, v1) = (vss, vdd) if harness.in_direction == 'rise' else (vdd, vss)
        circuit.PieceWiseLinearVoltageSource('in', 'vin', circuit.gnd, values=[
            (0, v0),
            (t['data_edge_1_start'], v0),
            (t['data_edge_1_end'], v1),
            (t['data_edge_2_start'], v1),
            (t['data_edge_2_end'], v0),
            (t['sim_end'], v0)
        ])

        # Set up set and reset node
        if harness.reset:
            circuit.V('rin', 'vrin', circuit.gnd, vdd if harness.reset.state == '1' else vss)
        if harness.set:
            circuit.V('sin', 'vsin', circuit.gnd, vdd if harness.set.state == '1' else vss)

        # Initialize device under test subcircuit and wire up ports
        ports = self.definition().upper().split()[1:]
        connections = [ports.pop(0)]
        for port in ports:
            if port == harness.target_in_port.pin.name:
                connections.append('vin')
            elif port == harness.target_out_port.pin.name:
                connections.append('vout')
            elif port == settings.vdd.name.upper():
                connections.append('vdd_dyn')
            elif port == settings.vss.name.upper():
                connections.append('vss_dyn')
            elif port == harness.clock.pin.name:
                connections.append('vcin')
            elif self.reset and port == harness.reset.pin.name:
                connections.append('vrin')
            elif self.set and port == harness.set.pin.name:
                connections.append('vsin')
            elif port in [pin.pin.name for pin in harness.stable_in_ports]:
                for stable_port in harness.stable_in_ports:
                    if port == stable_port.pin.name:
                        if stable_port.state == '1':
                            connections.append('vhigh')
                        elif stable_port.state == '0':
                            connections.append('vlow')
                        else:
                            raise ValueError(f'Invalid state identified during simulation setup for port {port}: {state}')
            else:
                connections.append('wfloat0') # Float any unrecognized ports
        if len(connections) is not len(ports)+1:
            raise ValueError(f'Failed to match all ports identified in definition "{self.definition().strip()}"')
        circuit.X('dut', *connections)

        # Initialize simulator
        simulator = circuit.simulator(temperature=settings.temperature,
                                      nominal_temperature=settings.temperature,
                                      simulator=settings.simulator)
        return simulator, t

    def _measure_cell_delays(self, settings, harness, simulator, timings, debug_path):
        """Run delay measurement for a single test circuit."""

        # Set up voltage bounds for measurements
        pct_vdd = lambda x : x * settings.vdd.voltage
        match harness.in_direction:
            case 'rise':
                v_prop_start = settings.logic_threshold_low_to_high
            case 'fall':
                v_prop_start = settings.logic_threshold_high_to_low
        match harness.out_direction:
            case 'rise':
                v_prop_end = settings.logic_threshold_low_to_high
                v_trans_start = settings.logic_threshold_low
                v_trans_end = settings.logic_threshold_high
            case 'fall':
                v_prop_end = settings.logic_threshold_high_to_low
                v_trans_start = settings.logic_threshold_high
                v_trans_end = settings.logic_threshold_low
        match harness.timing_type_clock:
            case 'rising_edge':
                clk_direction = 'rise'
                v_clk_transition = settings.logic_threshold_low_to_high
            case 'falling_edge':
                clk_direction = 'fall'
                v_clk_transition = settings.logic_threshold_high_to_low

        # Measure propagation delay from first data edge to last output edge
        simulator.measure('tran', 'prop_in_out',
            f'trig v(vin) val={pct_vdd(v_prop_start)} td={float(timings["removal"])} {harness.in_direction}=1',
            f'targ v(vout) val={pct_vdd(v_prop_end)} {harness.out_direction}=last')

        # Measure transient delay from first data edge to first output edge
        simulator.measure('tran', 'trans_out',
            f'trig v(vout) val={pct_vdd(v_trans_start)} td={float(timings["removal"])} {harness.out_direction}=1',
            f'targ v(vout) val={pct_vdd(v_trans_end)} {harness.out_direction}=1')

        # Measure setup delay from first data edge to last clock edge
        simulator.measure('tran', 't_setup',
            f'trig v(vin) val={pct_vdd(v_prop_start)} td={float(timings["removal"])} {harness.in_direction}=1',
            f'targ v(vcin) val={pct_vdd(v_clk_transition)} {clk_direction}=last')

        # Measure hold delay from last clock edge to last data edge
        simulator.measure('tran', 't_hold',
            f'trig v(vcin) val={pct_vdd(v_clk_transition)} td={float(timings["removal"])} {clk_direction}=last',
            f'targ v(vin) val={pct_vdd(v_prop_end)} {_flip_direction(harness.in_direction)}=last')

        # Log simulation
        if settings.debug:
            debug_path.mkdir(parents=True, exist_ok=True)
            with open(debug_path/'delay.sp', 'w') as spice_file:
                spice_file.write(str(simulator))

        step_time = min(self.sim_timestep*settings.units.time, timings['sim_end']/1000)
        simulator.options('autostop', 'nopage', 'nomod', post=1, ingold=2)
        return simulator.transient(step_time=step_time, end_time=timings['sim_end'])

    def _measure_c2q(self, settings, harness, simulator, timings, debug_path):
        """Measure Clock-to-Q delay."""

        # Measure clock-to-latch time
        match harness.timing_type_clock:
            case 'rising_edge':
                clk_direction = 'rise'
                v_clk_transition = settings.logic_threshold_low_to_high * settings.vdd.voltage
            case 'falling_edge':
                clk_direction = 'fall'
                v_clk_transition = settings.logic_threshold_high_to_low * settings.vdd.voltage
        match harness.out_direction:
            case 'rise':
                v_prop_end = settings.logic_threshold_low_to_high * settings.vdd.voltage
            case 'fall':
                v_prop_end = settings.logic_threshold_high_to_low * settings.vdd.voltage
        simulator.measure('tran', 't_c2q',
            f'trig v(vcin) val={v_clk_transition} td={float(timings["removal"])} {clk_direction}=last',
            f'targ v(vout) val={v_prop_end} {harness.out_direction}=last')

        # Log simulation
        if settings.debug:
            debug_path = debug_path / 'c2q'
            debug_path.mkdir(parents=True, exist_ok=True)
            with open(debug_path/f'{simulator.circuit.title}.sp', 'w') as spice_file:
                spice_file.write(str(simulator))

        step_time = min(self.sim_timestep*settings.units.time, t['sim_end']/1000)
        simulator.options('autostop', 'nopage', 'nomod', post=1, ingold=2)
        return simulator.transient(step_time=step_time, end_time=timings['sim_end'])

    def plot_io(self, settings, harness):
        """Plot I/O voltages vs time"""
        # TODO: Look for ways to generate fewer plots here - maybe a creative 3D plot
        figures = []
        # Group data by slew rate so that inputs are the same
        for slew in self.in_slews:
            for load in self.out_loads:
                # Add axes for clk, s, r, d, q (in that order)
                # Use an additive approach in case some of those aren't present
                num_axes = 1
                CLK = 0
                if self.set:
                    S = num_axes
                    num_axes += 1
                if self.reset:
                    R = num_axes
                    num_axes += 1
                D = num_axes
                num_axes += 1
                Q = num_axes
                num_axes += 1
                ratios = np.ones(num_axes).tolist()
                ratios[-1] = num_axes
                figure, axes = plt.subplots(num_axes,
                    sharex=True,
                    height_ratios=ratios,
                    label=f'{self.cell.name} | {harness.short_str()}'
                )

                # Set up plots
                for ax in axes:
                    for level in [settings.logic_threshold_low, settings.logic_threshold_high]:
                        ax.axhline(level*settings.vdd.voltage, color='0.5', linestyle='--')
                    # TODO: Set up vlines for important timing events
                    ax.set_yticks([settings.vss.voltage, settings.vdd.voltage])
                volt_units = str(settings.units.voltage.prefixed_unit)
                time_units = str(settings.units.time.prefixed_unit)
                axes[CLK].set(
                    title=f'Slew Rate: {str(slew*settings.units.time)} | Fanout: {str(load*settings.units.capacitance)}',
                    ylabel=f'CLK [{volt_units}]'
                )
                if self.set:
                    axes[S].set_ylabel(f'S [{volt_units}]')
                if self.reset:
                    axes[R].set_ylabel(f'R [{volt_units}]')
                axes[D].set_ylabel(f'D [{volt_units}]')
                axes[Q].set_ylabel(f'Q [{volt_units}]')
                axes[-1].set_xlabel(f'Time [{str(settings.units.time.prefixed_unit)}]')
                analysis = harness.results[str(slew)][str(load)]
                t = analysis.time / settings.units.time
                axes[CLK].plot(t, analysis.vcin)
                if self.set:
                    axes[S].plot(t, analysis.vsin)
                if self.reset:
                    axes[R].plot(t, analysis.vrin)
                axes[D].plot(t, analysis.vin)
                axes[Q].plot(t, analysis.vout)

                figures.append(figure)
        return figures

def _flip_direction(direction: str) -> str:
    return 'fall' if direction == 'rise' else 'rise'

def _gen_graycode(length: int):
    """Generate the list of Gray Codes of specified length"""
    if length <= 1:
        return [[0],[1]]
    inputs = []
    for j in _gen_graycode(length-1):
        j.insert(0, 0)
        inputs.append(j)
    for j in reversed(_gen_graycode(length-1)):
        j.insert(0, 1)
        inputs.append(j)
    return inputs

def _parse_triggered_pin(value: str, role: str) -> (str, Pin):
    """Parses input pin names with trigger types, e.g. 'posedge CLK'"""
    if not isinstance(value, str):
        raise TypeError(f'Invalid type for edge-triggered pin: {type(value)}')
    try:
        edge, name = value.split()
    except ValueError:
        raise ValueError(f'Invalid value for edge-triggered pin: {value}. Make sure you include both the trigger type and pin name (e.g. "posedge CLK")')
    if not edge in ['posedge', 'negedge']:
        raise ValueError(f'Invalid trigger type: {edge}. Trigger type must be one of "posedge" or "negedge"')
    return (edge, Pin(name, 'input', role))
